"""
Module for the PrivateTab

A PrivateTab is a private conversation opened with someone from a MUC
(see muctab.py). The conversation happens with both JID being relative
to the MUC (room@server/nick1 and room@server/nick2).

This tab references his parent room, and is modified to keep track of
both participant’s nicks. It also has slightly different features than
the ConversationTab (such as tab-completion on nicks from the room).

"""
import curses
import logging
from datetime import datetime
from typing import Any, Callable, TYPE_CHECKING

from slixmpp import JID
from slixmpp.stanza import Message as SMessage

from poezio.tabs import OneToOneTab, MucTab, Tab

from poezio import common
from poezio import windows
from poezio import xhtml
from poezio.config import config, get_image_cache
from poezio.core.structs import Command
from poezio.decorators import refresh_wrapper
from poezio.theming import get_theme, dump_tuple
from poezio.decorators import command_args_parser
from poezio.text_buffer import CorrectionError
from poezio.ui.types import (
    Message,
    PersistentInfoMessage,
)

if TYPE_CHECKING:
    from poezio.core.core import Core
    from poezio.user import User

log = logging.getLogger(__name__)


class PrivateTab(OneToOneTab):
    """
    The tab containing a private conversation (someone from a MUC)
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}
    message_type = 'chat'
    parent_muc: MucTab | None
    additional_information: dict[str, Callable[[str], str]] = {}

    def __init__(self, core: 'Core', jid: JID, nick: str,
                 muctab: MucTab | None, initial:  SMessage | None = None) -> None:
        OneToOneTab.__init__(self, core, jid, initial)
        self.own_nick = nick
        self.info_header = windows.PrivateInfoWin()
        # keys
        self.key_func['^I'] = self.completion
        # commands
        self.register_command(
            'info',
            self.command_info,
            desc=(
                'Display some information about the user in the MUC: '
                'their role, affiliation, status and status message.'
            ),
            shortdesc='Info about the user.')
        self.register_command(
            'version',
            self.command_version,
            desc=(
                'Get the software version of the current interlocutor '
                '(usually their XMPP client and Operating System).'
            ),
            shortdesc='Get the software version of a jid.')
        self.resize()
        self.parent_muc = muctab
        self.on = True
        self.update_commands()
        self.update_keys()

    @property
    def log_name(self) -> str:
        """Overridden from ChatTab because this is a case where we want the full JID"""
        return self.jid.full

    def remote_user_color(self) -> str:
        if self.parent_muc:
            user = self.parent_muc.get_user_by_name(self.jid.resource)
            if user:
                return dump_tuple(user.color)
        return super().remote_user_color()

    @property
    def general_jid(self) -> JID:
        return self.jid

    def get_dest_jid(self) -> JID:
        return self.jid

    @property
    def nick(self) -> str:
        return self.get_nick()

    @staticmethod
    @refresh_wrapper.always
    def add_information_element(plugin_name: str, callback: Callable[[str], str]) -> None:
        """
        Lets a plugin add its own information to the PrivateInfoWin
        """
        PrivateTab.additional_information[plugin_name] = callback

    @staticmethod
    @refresh_wrapper.always
    def remove_information_element(plugin_name: str) -> None:
        del PrivateTab.additional_information[plugin_name]

    def on_close(self) -> None:
        super().on_close()
        if self.parent_muc:
            self.parent_muc.privates.remove(self)

    def completion(self) -> None:
        """
        Called when Tab is pressed, complete the nickname in the input
        """
        if self.complete_commands(self.input):
            return

        # If we are not completing a command or a command's argument, complete a nick
        def compare_users(user: User) -> datetime:
            return user.last_talked
        # No parent MUC, unable to complete other user nickames
        if not self.parent_muc:
            return
        word_list = [
            user.nick for user in sorted(self.parent_muc.users, key=compare_users, reverse=True)
            if user.nick != self.own_nick
        ]
        after = config.getstr('after_completion') + ' '
        input_pos = self.input.pos
        if ' ' not in self.input.get_text()[:input_pos] or \
                (self.input.last_completion and
                 self.input.get_text()[:input_pos] == self.input.last_completion + after):
            add_after = after
        else:
            add_after = ''
        self.input.auto_completion(word_list, add_after, quotify=False)
        empty_after = self.input.get_text() == '' or (
            self.input.get_text().startswith('/')
            and not self.input.get_text().startswith('//'))
        self.send_composing_chat_state(empty_after)

    async def handle_message(self, msg: SMessage, display: bool = True) -> None:
        sent = msg['from'].bare == self.core.xmpp.boundjid.bare
        jid = msg['to'] if sent else msg['from']
        with_nick = jid.resource
        sender_nick = with_nick
        if sent:
            sender_nick = (self.own_nick or self.core.own_nick)
        room_from = jid.bare
        use_xhtml = config.get_by_tabname(
            'enable_xhtml_im',
            jid.bare
        )
        tmp_dir = get_image_cache()
        if not sent:
            await self.core.events.trigger_async('private_msg', msg, self)
        body = xhtml.get_body_from_message_stanza(
            msg, use_xhtml=use_xhtml, extract_images_to=tmp_dir)
        if not body or not self:
            return
        delayed, date = common.find_delayed_tag(msg)
        replaced = False
        user = None
        if self.parent_muc:
            user = self.parent_muc.get_user_by_name(with_nick)
        if msg.get_plugin('replace', check=True):
            replaced_id = msg['replace']['id']
            if replaced_id != '' and config.get_by_tabname(
                    'group_corrections', room_from):
                try:
                    self.modify_message(
                        body,
                        replaced_id,
                        msg['id'],
                        user=user,
                        time=date,
                        jid=msg['from'],
                        nickname=sender_nick)
                    replaced = True
                except CorrectionError:
                    log.debug('Unable to correct a message', exc_info=True)
        if not replaced:
            message = Message(
                txt=body,
                time=date,
                history=delayed,
                nickname=sender_nick,
                nick_color=get_theme().COLOR_OWN_NICK if sent else None,
                user=user,
                identifier=msg['id'],
                jid=msg['from'],
            )
            if display:
                self.add_message(message)
            else:
                self.log_message(message)
        if sent:
            self.set_last_sent_message(msg, correct=replaced)
        else:
            self.last_remote_message = datetime.now()

    @refresh_wrapper.always
    @command_args_parser.raw
    async def command_say(self, line: str, attention: bool = False, correct: bool = False) -> None:
        if not self.on:
            return
        await self._initial_log.wait()
        msg: SMessage = self.core.xmpp.make_message(
            mto=self.jid.full,
            mfrom=self.core.xmpp.boundjid
        )
        msg['type'] = 'chat'
        msg['body'] = line
        msg.enable('muc')
        # trigger the event BEFORE looking for colors.
        # This lets a plugin insert \x19xxx} colors, that will
        # be converted in xhtml.
        self.core.events.trigger('private_say', msg, self)
        if not msg['body']:
            return
        if correct or msg['replace']['id'] and self.last_sent_message:
            msg['replace']['id'] = self.last_sent_message['id']  # type: ignore
        else:
            del msg['replace']

        if msg['body'].find('\x19') != -1:
            msg.enable('html')
            msg['html']['body'] = xhtml.poezio_colors_to_html(msg['body'])
            msg['body'] = xhtml.clean_text(msg['body'])
        if config.get_by_tabname('send_chat_states', self.general_jid):
            if self.inactive:
                self.send_chat_state('inactive', always_send=True)
            else:
                msg['chat_state'] = 'active'
        if attention:
            msg['attention'] = True
        self.core.events.trigger('private_say_after', msg, self)
        if not msg['body']:
            return
        self.set_last_sent_message(msg, correct=correct)
        await self.core.handler.on_groupchat_private_message(msg, sent=True)
        # Our receipts slixmpp hack
        msg._add_receipt = True  # type: ignore
        msg.send()
        self.cancel_paused_delay()

    @command_args_parser.quoted(0, 1)
    async def command_version(self, args: list[str]) -> None:
        """
        /version
        """
        if args:
            await self.core.command.version(args[0])
            return
        iq = await self.core.xmpp.plugin['xep_0092'].get_version(self.jid)
        self.core.handler.on_version_result(iq)

    @command_args_parser.quoted(0, 1)
    def command_info(self, arg: list[str]) -> None:
        """
        /info
        """
        if not self.parent_muc:
            return
        if arg and arg[0]:
            self.parent_muc.command_info(arg[0])
        else:
            user = self.jid.resource
            self.parent_muc.command_info(user)

    def resize(self) -> None:
        self.need_resize = False

        if self.size.tab_degrade_y:
            info_win_height = 0
            tab_win_height = 0
        else:
            info_win_height = self.core.information_win_size
            tab_win_height = Tab.tab_win_height()

        self.text_win.resize(
            self.height - 2 - info_win_height - tab_win_height, self.width, 0,
            0, self._text_buffer, force=self.ui_config_changed)
        self.ui_config_changed = False
        self.info_header.resize(
            1, self.width, self.height - 2 - info_win_height - tab_win_height,
            0)
        self.input.resize(1, self.width, self.height - 1, 0)

    def refresh(self) -> None:
        if self.need_resize:
            self.resize()
        log.debug('  TAB   Refresh: %s', self.__class__.__name__)
        display_info_win = not self.size.tab_degrade_y

        self.text_win.refresh()
        self.info_header.refresh(self.jid.full, self.text_win, self.chatstate,
                                 PrivateTab.additional_information)
        if display_info_win:
            self.info_win.refresh()

        self.refresh_tab_win()
        self.input.refresh()

    def refresh_info_header(self) -> None:
        self.info_header.refresh(self.jid.full, self.text_win, self.chatstate,
                                 PrivateTab.additional_information)
        self.input.refresh()

    def get_nick(self) -> str:
        return self.jid.resource

    def on_input(self, key: str, raw: bool) -> None:
        if not raw and key in self.key_func:
            self.key_func[key]()
            return
        self.input.do_command(key, raw=raw)
        if not self.on:
            return
        empty_after = self.input.get_text() == '' or (
            self.input.get_text().startswith('/')
            and not self.input.get_text().startswith('//'))
        tab = self.core.tabs.by_name_and_class(self.jid.bare, MucTab)
        if tab and tab.joined:
            self.send_composing_chat_state(empty_after)

    def on_lose_focus(self) -> None:
        if self.input.text:
            self.state = 'nonempty'
        else:
            self.state = 'normal'

        self.text_win.remove_line_separator()
        self.text_win.add_line_separator(self._text_buffer)
        tab = self.core.tabs.by_name_and_class(self.jid.bare, MucTab)
        if tab and tab.joined and config.get_by_tabname(
                'send_chat_states', self.general_jid) and self.on:
            self.send_chat_state('inactive')
        self.check_scrolled()

    def on_gain_focus(self) -> None:
        self.state = 'current'
        curses.curs_set(1)
        tab = self.core.tabs.by_name_and_class(self.jid.bare, MucTab)
        if tab and tab.joined and config.get_by_tabname(
                'send_chat_states',
                self.general_jid,
        ) and not self.input.get_text() and self.on:
            self.send_chat_state('active')

    def on_info_win_size_changed(self) -> None:
        if self.core.information_win_size >= self.height - 3:
            return
        self.text_win.resize(
            self.height - 2 - self.core.information_win_size -
            Tab.tab_win_height(), self.width, 0, 0)
        self.info_header.resize(
            1, self.width, self.height - 2 - self.core.information_win_size -
            Tab.tab_win_height(), 0)

    @refresh_wrapper.conditional
    def rename_user(self, old_nick: str, user: 'User') -> bool:
        """
        The user changed her nick in the corresponding muc: update the tab’s name and
        display a message.
        """
        self.add_message(
            PersistentInfoMessage(
                '\x19%(nick_col)s}%(old)s\x19%(info_col)s} is now '
                'known as \x19%(nick_col)s}%(new)s' % {
                    'old': old_nick,
                    'new': user.nick,
                    'nick_col': dump_tuple(user.color),
                    'info_col': dump_tuple(get_theme().COLOR_INFORMATION_TEXT)
                },
            ),
        )
        new_jid = self.jid.bare + '/' + user.nick
        self._name = new_jid
        return self.core.tabs.current_tab is self

    @refresh_wrapper.conditional
    def user_left(self, status_message: str, user: 'User') -> bool:
        """
        The user left the associated MUC
        """
        self.deactivate()
        hide = config.get_by_tabname('hide_exit_join', self.general_jid)
        if hide == 0:
            return self.core.tabs.current_tab is self
        theme = get_theme()
        if config.get_by_tabname('display_user_color_in_join_part',
                                 self.general_jid):
            color = dump_tuple(user.color)
        else:
            color = dump_tuple(theme.COLOR_REMOTE_USER)

        if not status_message:
            self.add_message(
                PersistentInfoMessage(
                    '\x19%(quit_col)s}%(spec)s \x19%(nick_col)s}'
                    '%(nick)s\x19%(info_col)s} has left the room' % {
                        'nick': user.nick,
                        'spec': theme.CHAR_QUIT,
                        'nick_col': color,
                        'quit_col': dump_tuple(theme.COLOR_QUIT_CHAR),
                        'info_col': dump_tuple(theme.COLOR_INFORMATION_TEXT)
                    },
                ),
            )
        else:
            self.add_message(
                PersistentInfoMessage(
                    '\x19%(quit_col)s}%(spec)s \x19%(nick_col)s}'
                    '%(nick)s\x19%(info_col)s} has left the room'
                    ' (%(status)s)' % {
                        'status': status_message,
                        'nick': user.nick,
                        'spec': theme.CHAR_QUIT,
                        'nick_col': color,
                        'quit_col': dump_tuple(theme.COLOR_QUIT_CHAR),
                        'info_col': dump_tuple(theme.COLOR_INFORMATION_TEXT)
                    },
                ),
            )
        return self.core.tabs.current_tab is self

    @refresh_wrapper.conditional
    def user_rejoined(self, nick: str) -> bool:
        """
        The user (or at least someone with the same nick) came back in the MUC
        """
        self.activate()
        hide = config.get_by_tabname('hide_exit_join', self.general_jid)
        if hide == 0:
            return self.core.tabs.current_tab is self
        muctab = self.parent_muc
        # We do not have a muctab for this user, we should not receive
        # presences from them;
        if muctab is None:
            return False
        theme = get_theme()
        color = dump_tuple(theme.COLOR_REMOTE_USER)
        if muctab and config.get_by_tabname('display_user_color_in_join_part',
                                            self.general_jid):
            user = muctab.get_user_by_name(nick)
            if user:
                color = dump_tuple(user.color)
        self.add_message(
            PersistentInfoMessage(
                '\x19%(join_col)s}%(spec)s \x19%(color)s}%(nick)s\x19'
                '%(info_col)s} joined the room' % {
                    'nick': nick,
                    'color': color,
                    'spec': theme.CHAR_JOIN,
                    'join_col': dump_tuple(theme.COLOR_JOIN_CHAR),
                    'info_col': dump_tuple(theme.COLOR_INFORMATION_TEXT)
                },
            ),
        )
        return self.core.tabs.current_tab is self

    def activate(self, reason: str | None = None) -> None:
        self.on = True
        if reason:
            self.add_message(PersistentInfoMessage(reason))

    def deactivate(self, reason: str | None = None) -> None:
        self.on = False
        if reason:
            self.add_message(PersistentInfoMessage(reason))

    def matching_names(self) -> list[tuple[int, str]]:
        return [(3, self.jid.resource), (4, self.name)]

    def add_error(self, error_message: str) -> None:
        theme = get_theme()
        error = '\x19%s}%s\x19o' % (dump_tuple(theme.COLOR_CHAR_NACK),
                                    error_message)
        self.add_message(
            Message(
                error,
                highlight=True,
                nickname='Error',
                nick_color=theme.COLOR_ERROR_MSG,
            ),
        )
        self.core.refresh_window()



from poezio.windows.base_wins import Win

from poezio.theming import get_theme, to_curses_attr


class Dialog(Win):
    __slots__ = ('text', 'accept', 'critical')

    str_accept = " Accept "
    str_refuse = " Reject "

    def __init__(self, helper_text: str, critical: bool = False) -> None:
        self.text = helper_text
        self.accept = False
        self.critical = critical

    def refresh(self) -> None:
        self._win.erase()
        self.addstr(self.text + "\n   ")

        if self.critical:
            col = to_curses_attr(get_theme().COLOR_WARNING_PROMPT)
        else:
            col = to_curses_attr(get_theme().COLOR_TAB_NORMAL)

        if self.accept:
            self.addstr(self.str_accept, col)
        else:
            self.addstr(self.str_accept)

        self.addstr("   ")

        if self.accept:
            self.addstr(self.str_refuse)
        else:
            self.addstr(self.str_refuse, col)

        self._refresh()

    def toggle_choice(self) -> None:
        self.accept = not self.accept



"""
Module exporting all the Windows, which are wrappers around curses wins
used to display information on the screen
"""

from poezio.windows.base_wins import Win
from poezio.windows.data_forms import FormWin
from poezio.windows.bookmark_forms import BookmarksWin
from poezio.windows.confirm import Dialog
from poezio.windows.info_bar import GlobalInfoBar, VerticalGlobalInfoBar
from poezio.windows.info_wins import InfoWin, XMLInfoWin, PrivateInfoWin, MucListInfoWin, \
        ConversationInfoWin, DynamicConversationInfoWin, MucInfoWin, \
        ConversationStatusMessageWin, BookmarksInfoWin, ConfirmStatusWin
from poezio.windows.input_placeholders import HelpText, NoInput
from poezio.windows.inputs import Input, HistoryInput, MessageInput, CommandInput, InputLike
from poezio.windows.list import ListWin, ColumnHeaderWin
from poezio.windows.misc import VerticalSeparator
from poezio.windows.muc import UserList, Topic
from poezio.windows.roster_win import RosterWin, ContactInfoWin
from poezio.windows.text_win import TextWin
from poezio.windows.image import ImageWin

__all__ = [
    'Win', 'FormWin', 'BookmarksWin', 'Dialog', 'GlobalInfoBar',
    'VerticalGlobalInfoBar', 'InfoWin', 'PrivateInfoWin', 'XMLInfoWin',
    'MucListInfoWin', 'ConversationInfoWin', 'MucInfoWin',
    'DynamicConversationInfoWin', 'ConversationStatusMessageWin',
    'BookmarksInfoWin', 'ConfirmStatusWin', 'HelpText', 'Input',
    'HistoryInput', 'MessageInput', 'CommandInput', 'ListWin',
    'ColumnHeaderWin', 'VerticalSeparator', 'UserList', 'Topic', 'RosterWin',
    'ContactInfoWin', 'TextWin', 'ImageWin', 'InputLike', 'NoInput',
]



"""
Shuffle the words in every message you send in a :ref:`muctab`
(may/should confuse the reader).
"""
from poezio.plugin import BasePlugin
from slixmpp import Message
from typing import Any
from random import shuffle
from poezio import xhtml


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.shuffle)

    def shuffle(self, msg: Message, tab: Any) -> None:
        split = xhtml.clean_text(msg['body']).split()
        shuffle(split)
        msg['body'] = ' '.join(split)



"""
Repeats the last message in the conversation.

Command
-------

.. glossary::

    /mirror
        **Usage:** ``/mirror``

"""
from poezio.plugin import BasePlugin
from poezio import tabs


class Plugin(BasePlugin):
    def init(self) -> None:
        for tab_type in (tabs.MucTab, tabs.PrivateTab, tabs.DynamicConversationTab, tabs.StaticConversationTab):
            self.api.add_tab_command(
                tab_type,
                'mirror',
                handler=self.mirror,
                help='Repeat the last message from the conversation.',
                short='Repeat the last message from the conversation.')

    def mirror(self, args: str) -> None:
        messages = self.api.get_conversation_messages()
        if not messages:
            # Do nothing if the conversation doesn’t contain any message
            return
        last_message = messages[-1]
        self.api.send_message(last_message.txt)



"""
Module related to the argument parsing
"""
import importlib.resources
import stat
import sys
from argparse import ArgumentParser, SUPPRESS, Namespace
from pathlib import Path
from shutil import copy2
from typing import Tuple

try:
    from poezio.version import __version__
except ImportError:
    # No version file, running from git without local install?
    __version__ = "dev"
from poezio import xdg


def parse_args(CONFIG_PATH: Path) -> Namespace:
    """
    Parse the arguments from the command line
    """
    parser = ArgumentParser('poezio')
    parser.add_argument(
        "-c",
        "--check-config",
        dest="check_config",
        action='store_true',
        help='Check the config file')
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        help="The file where debug will be written",
        metavar="DEBUG_FILE")
    parser.add_argument(
        "-f",
        "--file",
        dest="filename",
        default=CONFIG_PATH / 'poezio.cfg',
        type=Path,
        help="The config file you want to use",
        metavar="CONFIG_FILE")
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version='Poezio v%s' % __version__,
    )
    parser.add_argument(
        "--custom-version",
        dest="custom_version",
        help=SUPPRESS,
        metavar="VERSION",
        default=__version__
    )
    return parser.parse_args()


def run_cmdline_args() -> Tuple[Namespace, bool]:
    "Parse the command line arguments"
    options = parse_args(xdg.CONFIG_HOME)
    firstrun = False

    # Copy a default file if none exists
    if not options.filename.is_file():
        try:
            options.filename.parent.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            sys.stderr.write(
                'Poezio was unable to create the config directory: %s\n' % e)
            sys.exit(1)
        default = Path(__file__).parent / '..' / 'data' / 'default_config.cfg'
        try:
            other = importlib.resources.files('poezio') / 'default_config.cfg'
        except Exception:
            print(
                'Unable to locate the default poezio config file for initial setup.',
                file=sys.stderr,
            )
            other = Path()
        if default.is_file():
            copy2(str(default), str(options.filename))
        elif other.is_file():
            copy2(str(other), str(options.filename))
        else:
            print(
                'The poezio default config file does not seem to exist, there '
                'may be an issue in packaging.',
                file=sys.stderr,
            )

        # Inside the nixstore and possibly other distributions, the reference
        # file is readonly, so is the copy.
        # Make it writable by the user who just created it.
        if options.filename.exists():
            options.filename.chmod(options.filename.stat().st_mode
                                   | stat.S_IWUSR)
        firstrun = True

    return (options, firstrun)



"""
This plugin adds a message at 00:00 in each of your chat tabs saying that the
date has changed.

"""

import datetime
from gettext import gettext as _

from poezio import timed_events, tabs
from poezio.plugin import BasePlugin
from poezio.ui.types import InfoMessage


class Plugin(BasePlugin):
    next_event: timed_events.TimedEvent | None = None

    def init(self) -> None:
        self.schedule_event()

    def cleanup(self) -> None:
        if self.next_event is not None:
            self.api.remove_timed_event(self.next_event)

    def schedule_event(self) -> None:
        day_change = datetime.datetime.combine(datetime.date.today(),
                                               datetime.time())
        day_change += datetime.timedelta(1)
        self.next_event = timed_events.TimedEvent(day_change, self.day_change)
        self.api.add_timed_event(self.next_event)

    def day_change(self) -> None:
        msg = _("Day changed to %s") % (datetime.date.today().isoformat())

        for tab in self.core.tabs:
            if isinstance(tab, tabs.ChatTab):
                tab.add_message(InfoMessage(msg))

        self.core.refresh_window()
        self.schedule_event()



"""
This plugin will set your status to **away** if you detach your screen.

The default behaviour is to check for both tmux and screen (in that order).

.. note:: This plugins requires the pyinotify python package to load.

Configuration options
---------------------

.. glossary::

    use_screen
        **Default:** ``true``

        Try to find an attached screen.

    use_tmux
        **Default:** ``true``

        Try to find and attached tmux.

    use_csi
        **Default:** ``false``

        Use `client state indication`_ to limit bandwidth (thus CPU) usage when detached. WARNING: using CSI together with chatrooms will result in inaccurate logs due to presence filtering or other inaccuracies.

.. _client state indication: https://xmpp.org/extensions/xep-0352.html
"""

from poezio.plugin import BasePlugin
import os
import stat
import pyinotify
import asyncio

DEFAULT_CONFIG = {
    'screen_detach': {
        'use_tmux': True,
        'use_screen': True,
        'use_csi': False
    }
}

# overload if this is not how your stuff
# is configured
try:
    LOGIN = os.getlogin() or ''
    LOGIN_TMUX = os.getuid()
except Exception:
    LOGIN = os.getenv('USER') or ''
    LOGIN_TMUX = os.getuid()

SCREEN_DIR = '/var/run/screens/S-%s' % LOGIN
TMUX_DIR = '/tmp/tmux-%s' % LOGIN_TMUX


def find_screen(path: str) -> str | None:
    if not os.path.isdir(path):
        return None
    for f in os.listdir(path):
        path = os.path.join(path, f)
        if screen_attached(path):
            return path
    return None


def screen_attached(socket: str) -> bool:
    return (os.stat(socket).st_mode & stat.S_IXUSR) != 0


class Plugin(BasePlugin, pyinotify.Notifier):

    default_config = DEFAULT_CONFIG

    def init(self) -> None:
        sock_path = None
        if self.config.get('use_tmux'):
            sock_path = find_screen(TMUX_DIR)
        if sock_path is None and self.config.get('use_screen'):
            sock_path = find_screen(SCREEN_DIR)

        # Only actually do something if we found an attached screen (assuming only one)
        if sock_path:
            self.attached = True
            wm = pyinotify.WatchManager()
            wm.add_watch(sock_path,
                         pyinotify.EventsCodes.ALL_FLAGS['IN_ATTRIB'])
            pyinotify.Notifier.__init__(
                self, wm, default_proc_fun=HandleScreen(plugin=self))
            asyncio.get_event_loop().add_reader(self._fd, self.process)
        else:
            self.api.information(
                'screen_detach plugin: No tmux or screen found', 'Warning')
            self.attached = False

    def process(self) -> None:
        self.read_events()
        self.process_events()

    def cleanup(self) -> None:
        asyncio.get_event_loop().remove_reader(self._fd)

    def update_screen_state(self, socket: str) -> None:
        attached = screen_attached(socket)
        if attached != self.attached:
            self.attached = attached
            status = 'available' if self.attached else 'away'
            self.core.command.status(status)
            if self.config.get('use_csi'):
                if self.attached:
                    self.core.xmpp.plugin['xep_0352'].send_active()
                else:
                    self.core.xmpp.plugin['xep_0352'].send_inactive()


class HandleScreen(pyinotify.ProcessEvent):
    def my_init(self, **kwargs) -> None:
        self.plugin = kwargs['plugin']

    def process_IN_ATTRIB(self, event) -> None:
        self.plugin.update_screen_state(event.path)



"""
XMPP-related handlers for the Core class
"""

import logging

import asyncio
import curses
import select
import signal
import ssl
import sys
import time
from hashlib import sha1, sha256, sha512
from typing import Any, TYPE_CHECKING

import pyasn1.codec.der.decoder
import pyasn1.codec.der.encoder
import pyasn1_modules.rfc2459
from slixmpp import InvalidJID, JID, Message, Iq, Presence
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.stanza.stream_error import StreamError
from slixmpp.xmlstream.stanzabase import StanzaBase, ElementBase
from slixmpp.plugins.xep_0410 import PingStatus
from slixmpp.plugins.xep_0004 import Form
from xml.etree import ElementTree as ET

from poezio import tabs
from poezio import xhtml
from poezio import multiuserchat as muc
from poezio.common import get_error_message
from poezio.config import config, get_image_cache
from poezio.core.structs import Status
from poezio.contact import Resource
from poezio.logger import logger
from poezio.roster import roster
from poezio.text_buffer import AckError
from poezio.theming import dump_tuple, get_theme
from poezio.ui.types import (
    XMLLog,
    InfoMessage,
    PersistentInfoMessage,
)

from poezio.core.commands import dumb_callback

try:
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters import HtmlFormatter
    LEXER = get_lexer_by_name('xml')
    FORMATTER = HtmlFormatter(noclasses=True)
    PYGMENTS = True
except ImportError:
    PYGMENTS = False


if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)

CERT_WARNING_TEXT = """
WARNING: CERTIFICATE FOR %s CHANGED

This can be part of a normal renewal process, but can also mean that \
an attacker is performing a man-in-the-middle attack on your connection.
When in doubt, check with your administrator using another channel.

SHA-256 of the old certificate (SPKI): %s

SHA-256 of the new certificate (SPKI): %s
"""

HTTP_VERIF_TEXT = """
Someone (maybe you) has requested an identity verification
using method "%s" for the url "%s".

The transaction id is: %s
And the XMPP address of the verification service is %s.

"""


class HandlerCore:
    def __init__(self, core: 'Core') -> None:
        self.core = core

    async def on_session_start_features(self, _: Any) -> None:
        """
        Enable carbons & blocking on session start if wanted and possible
        """
        iq = await self.core.xmpp.plugin['xep_0030'].get_info(
            jid=JID(self.core.xmpp.boundjid.domain),
        )
        features = iq['disco_info']['features']

        rostertab = self.core.tabs.by_name_and_class(
            'Roster', tabs.RosterInfoTab)
        if rostertab is not None:
            rostertab.check_saslexternal(features)
            rostertab.check_blocking(features)
        self.core.check_blocking(features)
        if (config.getbool('enable_carbons')
                and 'urn:xmpp:carbons:2' in features):
            self.core.xmpp.plugin['xep_0280'].enable()
        await self.core.check_bookmark_storage(features)

    def find_identities(self, _: Any) -> None:
        self.core.register_task(
            self.core.xmpp['xep_0030'].get_info_from_domain(),
        )

    def is_known_muc_pm(self, message: Message, with_jid: JID) -> bool | None:
        """
        Try to determine whether a given message is a MUC-PM
        without a roundtrip. Returns None when it's not clear.
        """

        # first, look for the x (XEP-0045 version 1.28)
        if message.match('message/muc'):
            log.debug('MUC-PM from %s with <x>', with_jid)
            return True

        jid_bare = with_jid.bare

        # then, look whether we have a matching tab with barejid
        tab = self.core.tabs.by_jid(JID(jid_bare))
        if tab is not None:
            if isinstance(tab, tabs.MucTab):
                log.debug('MUC-PM from %s in known MucTab', with_jid)
                return True
            one_to_one = isinstance(tab, (
                tabs.ConversationTab,
                tabs.DynamicConversationTab,
            ))
            if one_to_one:
                return False

        # then, look whether we have a matching tab with fulljid
        if with_jid.resource:
            tab = self.core.tabs.by_jid(with_jid)
            if tab is not None:
                if isinstance(tab, tabs.PrivateTab):
                    log.debug('MUC-PM from %s in known PrivateTab', with_jid)
                    return True
                if isinstance(tab, tabs.StaticConversationTab):
                    return False

        # then, look in the roster
        if jid_bare in roster and roster.contacts[jid_bare].subscription != 'none':
            return False

        # then, check bookmarks
        for bm in self.core.bookmarks:
            if bm.jid.bare == jid_bare:
                log.debug('MUC-PM from %s in bookmarks', with_jid)
                return True

        return None

    def muc_ping_result(self, result: dict) -> None:
        """The slix 410 plugin got a new result"""
        room_from, _ = result.get('key', (None, None))
        room_from = JID(room_from.bare)
        tab = self.core.tabs.by_name_and_class(room_from, tabs.MucTab)
        if tab:
            prev: PingStatus = result.get('previous', PingStatus.UNTRIED)
            new: PingStatus = result.get('result', PingStatus.UNTRIED)
            if new != prev:
                tab.on_self_ping_result(new)

    async def on_carbon_received(self, message: Message) -> None:
        """
        Carbon <received/> received
        """
        recv = message['carbon_received']
        is_muc_pm = is_muc = self.is_known_muc_pm(recv, recv['from'])
        if is_muc_pm:
            if recv.match('message/muc/invite'):
                log.debug('Detected a MUC invitation.')
                await self.on_groupchat_invitation(recv)
            else:
                log.debug('%s sent a MUC-PM, ignoring carbon', recv['from'])
        elif is_muc_pm is None:
            is_muc = await self.core.xmpp.plugin['xep_0030'].has_identity(
                recv['from'].bare,
                node='conference',
            )
            if is_muc:
                log.debug('%s has category conference, ignoring carbon',
                          recv['from'].server)
        if not is_muc:
            recv['to'] = self.core.xmpp.boundjid.full
            if 'receipt' in recv:
                await self.on_receipt(recv)
            else:
                await self.on_normal_message(recv)

    async def on_carbon_sent(self, message: Message) -> None:
        """
        Carbon <sent/> received
        """
        sent = message['carbon_sent']
        is_muc_pm = self.is_known_muc_pm(sent, sent['to'])
        if is_muc_pm:
            await self.on_groupchat_private_message(sent, sent=True)
        elif is_muc_pm is None:
            is_muc = await self.core.xmpp.plugin['xep_0030'].has_identity(
                sent['to'].bare,
                node='conference',
            )
            if is_muc:
                await self.on_groupchat_private_message(sent, sent=True)
            else:
                sent['from'] = self.core.xmpp.boundjid.full
                await self.on_normal_message(sent)
        else:
            sent['from'] = self.core.xmpp.boundjid.full
            await self.on_normal_message(sent)

    ### Invites ###

    async def on_groupchat_invitation(self, message: Message) -> None:
        """
        Mediated invitation received
        """
        jid = message['from']
        if jid.bare in self.core.pending_invites:
            return
        invite = message['muc']['invite']
        # TODO: find out why pylint thinks "inviter" is a list
        # pylint: disable=no-member
        inviter = invite['from']
        reason = invite['reason']
        password = invite['password']
        msg = "You are invited to the room %s by %s" % (jid.full, inviter.full)
        if reason:
            msg += "because: %s" % reason
        if password:
            msg += ". The password is \"%s\"." % password
        self.core.information(msg, 'Info')
        if 'invite' in config.getstr('beep_on').split():
            curses.beep()
        logger.log_roster_change(inviter.full, 'invited you to %s' % jid.full)
        self.core.pending_invites[jid.bare] = inviter.full

    async def on_groupchat_decline(self, _: Any) -> None:
        "Mediated invitation declined; skip for now"
        pass

    async def on_groupchat_direct_invitation(self, message: Message) -> None:
        """
        Direct invitation received
        """
        try:
            room = JID(message['groupchat_invite']['jid'])
        except InvalidJID:
            return
        if room.bare in self.core.pending_invites:
            return

        inviter = message['from']
        reason = message['groupchat_invite']['reason']
        password = message['groupchat_invite']['password']
        continue_ = message['groupchat_invite']['continue']
        msg = "You are invited to the room %s by %s" % (room, inviter.full)

        if password:
            msg += ' (password: "%s")' % password
        if continue_:
            msg += '\nto continue the discussion'
        if reason:
            msg += "\nreason: %s" % reason

        self.core.information(msg, 'Info')
        if 'invite' in config.getstr('beep_on').split():
            curses.beep()

        self.core.pending_invites[room.bare] = inviter.full
        logger.log_roster_change(inviter.full, 'invited you to %s' % room.bare)

    ### "classic" messages ###

    async def on_message(self, message: Message) -> None:
        """
        When receiving private message from a muc OR a normal message
        (from one of our contacts)
        """
        if message.match('message/muc/invite'):
            return
        if message.match('message/retract'):
            return
        if message['type'] == 'groupchat':
            return
        # Differentiate both type of messages, and call the appropriate handler.
        if self.is_known_muc_pm(message, message['from']):
            await self.on_groupchat_private_message(message, sent=False)
        else:
            await self.on_normal_message(message)

    async def on_encrypted_message(self, message: Message) -> None:
        """
        When receiving an encrypted message
        """
        if message["body"]:
            return  # Already being handled by on_message.
        await self.on_message(message)

    async def on_error_message(self, message: Message) -> None:
        """
        When receiving any message with type="error"
        """
        jid_from = message['from']
        for muc_tab in self.core.get_tabs(tabs.MucTab):
            if muc_tab.jid.bare == jid_from.bare:
                if jid_from.full == jid_from.bare:
                    self.core.room_error(message, jid_from.bare)
                else:
                    text = get_error_message(message)
                    p_tab = self.core.tabs.by_name_and_class(
                        jid_from.full, tabs.PrivateTab)
                    if p_tab:
                        p_tab.add_error(text)
                    else:
                        self.core.information(text, 'Error')
                return
        tab = self.core.get_conversation_by_jid(message['from'], create=False)
        error_msg = get_error_message(message, deprecated=True)
        if not tab:
            self.core.information(error_msg, 'Error')
            return
        error = '\x19%s}%s\x19o' % (dump_tuple(get_theme().COLOR_CHAR_NACK),
                                    error_msg)
        if not tab.nack_message('\n' + error, message['id'], message['to']):
            tab.add_message(InfoMessage(error))
            self.core.refresh_window()

    async def on_message_retract(self, message: Message) -> None:
        jid_from = message['from']
        tab = (
            self.core.tabs.by_name_and_class(
                jid_from.full, tabs.ChatTab
            ) or self.core.tabs.by_name_and_class(
                jid_from.bare, tabs.ChatTab
            )
        )
        if tab is not None:
            tab.handle_retract(message)

    def on_reactions(self, message: Message, sent: bool = False) -> None:
        """
        When receiving reactions as a standalone message (this does not
        trigger the "message" or "groupchat_message" slixmpp events as there
        is usually no body alongside the reactions.
        """
        tab: tabs.MucTab | tabs.ConversationTab | tabs.PrivateTab | None
        if message['type'] == 'groupchat':
            tab = self.core.tabs.by_name_and_class(message['from'].bare, tabs.MucTab)
        elif self.is_known_muc_pm(message, message['from']):
            jid = message['to'] if sent else message['from']
            tab = self.core.tabs.by_name_and_class(
                jid.full,
                tabs.PrivateTab,
            )
        else:
            if message['to'].bare == self.core.xmpp.boundjid.bare:
                jid = message['from']
            # we wrote the message (happens with carbons)
            elif message['from'].bare == self.core.xmpp.boundjid.bare:
                jid = message['to']
            tab = self.core.get_conversation_by_jid(jid, create=False)
        if tab:
            tab.handle_reactions(message)

    async def on_normal_message(self, message: Message) -> None:
        """
        When receiving "normal" messages (not a private message from a
        muc participant)
        """
        if message['to'].bare == self.core.xmpp.boundjid.bare:
            conv_jid = message['from']
            own = False
        # we wrote the message (happens with carbons)
        elif message['from'].bare == self.core.xmpp.boundjid.bare:
            conv_jid = message['to']
            own = True
        # we are not part of that message, drop it
        else:
            return

        if message['type'] == 'error':
            return
        elif message['type'] == 'headline' and message['body']:
            self.core.information(
                '%s says: %s' % (message['from'], message['body']),
                'Headline',
            )
            return

        conversation = self.core.get_conversation_by_jid(conv_jid, create=False)

        if conversation and 'reactions' in message:
            conversation.handle_reactions(message)

        use_xhtml = config.get_by_tabname('enable_xhtml_im',
                                          message['from'].bare)
        tmp_dir = get_image_cache()
        if not xhtml.get_body_from_message_stanza(
                message, use_xhtml=use_xhtml, extract_images_to=tmp_dir):
            if not self.core.xmpp.plugin['xep_0380'].has_eme(message):
                return
            self.core.xmpp.plugin['xep_0380'].replace_body_with_eme(message)

        # normal message, we are the recipient

        if conversation is None:
            conversation = tabs.DynamicConversationTab(
                self.core,
                JID(conv_jid.bare),
                initial=message,
            )
            self.core.tabs.append(conversation)
        else:
            await conversation.handle_message(message)

        if not own and 'private' in config.getstr('beep_on').split():
            if not config.get_by_tabname('disable_beep', conv_jid.bare):
                curses.beep()
        if self.core.tabs.current_tab is not conversation:
            if not own:
                conversation.state = 'private'
                self.core.refresh_tab_win()
            else:
                conversation.set_state('normal')
                self.core.refresh_tab_win()
        else:
            self.core.refresh_window()

    async def on_0084_avatar(self, msg: Message) -> None:
        jid = msg['from'].bare
        contact = roster[jid]
        if not contact:
            return
        log.debug('Received 0084 avatar update from %s', jid)
        try:
            metadata = msg['pubsub_event']['items']['item']['avatar_metadata'][
                'items']
        except Exception:
            log.debug('Failed getting metadata from 0084:', exc_info=True)
            return
        for info in metadata:
            avatar_hash = info['id']

            # First check whether we have it in cache.
            cached_avatar = self.core.avatar_cache.retrieve_by_jid(
                jid, avatar_hash)
            if cached_avatar:
                contact.avatar = cached_avatar
                log.debug('Using cached avatar for %s', jid)
                return

            # If we didn’t have any, query the data instead.
            if not info['url']:
                try:
                    result = await self.core.xmpp['xep_0084'].retrieve_avatar(
                        jid, avatar_hash, timeout=60)
                    avatar = result['pubsub']['items']['item']['avatar_data'][
                        'value']
                    if sha1(avatar).hexdigest().lower() != avatar_hash.lower():
                        log.debug(
                            'Avatar sha1 doesn’t match 0153 hash. (%s != %s)',
                            sha1(avatar).hexdigest().lower(),
                            avatar_hash.lower(),
                        )
                        continue
                    contact.avatar = avatar
                except Exception:
                    log.debug(
                        'Failed retrieving 0084 data from %s:',
                        jid,
                        exc_info=True)
                    continue
                log.debug('Received %s avatar: %s', jid, info['type'])

                # Now we save the data on the file system to not have to request it again.
                if not self.core.avatar_cache.store_by_jid(
                        jid, avatar_hash, contact.avatar):
                    log.debug(
                        'Failed writing %s’s avatar to cache:',
                        jid,
                        exc_info=True)
                return

    async def on_vcard_avatar(self, pres: Presence) -> None:
        jid = pres['from'].bare
        contact = roster[jid]
        if not contact:
            return
        avatar_hash = pres['vcard_temp_update']['photo']
        log.debug('Received vCard avatar update from %s: %s', jid, avatar_hash)

        # First check whether we have it in cache.
        cached_avatar = self.core.avatar_cache.retrieve_by_jid(
            jid, avatar_hash)
        if cached_avatar:
            contact.avatar = cached_avatar
            log.debug('Using cached avatar for %s', jid)
            return

        # If we didn’t have any, query the vCard instead.
        try:
            result = await self.core.xmpp['xep_0054'].get_vcard(
                jid, cached=True, timeout=60)
            avatar = result['vcard_temp']['PHOTO']
            binval = avatar['BINVAL']
            if sha1(binval).hexdigest().lower() != avatar_hash.lower():
                log.debug(
                    'Avatar sha1 doesn’t match 0153 hash. (%s != %s)',
                    sha1(binval).hexdigest().lower(), avatar_hash.lower(),
                )
                return
            contact.avatar = binval
        except Exception:
            log.debug('Failed retrieving vCard from %s:', jid, exc_info=True)
            return
        log.debug('Received %s avatar: %s', jid, avatar['TYPE'])

        # Now we save the data on the file system to not have to request it again.
        if not self.core.avatar_cache.store_by_jid(jid, avatar_hash,
                                                   contact.avatar):
            log.debug(
                'Failed writing %s’s avatar to cache:', jid, exc_info=True)

    async def on_nick_received(self, message: Message) -> None:
        """
        Called when a pep notification for a user nickname
        is received
        """
        contact = roster[message['from'].bare]
        if not contact:
            return
        item = message['pubsub_event']['items']['item']
        if item.xml.find('{http://jabber.org/protocol/nick}nick') is not None:
            contact.name = item['nick']['nick']
        else:
            contact.name = ''

    async def on_groupchat_message(self, message: Message) -> None:
        """
        Triggered whenever a message is received from a multi-user chat room.
        """
        if message.match('message/retract'):
            return
        room_from = message['from'].bare

        if message['type'] == 'error':  # Check if it's an error
            self.core.room_error(message, room_from)
            return

        tab = self.core.tabs.by_name_and_class(room_from, tabs.MucTab)
        if not tab:
            self.core.information(
                "message received for a non-existing room: %s" % (room_from),
                'Error',
            )
            muc.leave_groupchat(
                self.core.xmpp, room_from, self.core.own_nick, msg='')
            return
        valid_message = await tab.handle_message(message)
        if valid_message and 'message' in config.getstr('beep_on').split():
            if (not config.get_by_tabname('disable_beep', room_from)
                    and self.core.own_nick != message['from'].resource):
                curses.beep()

    def on_muc_own_nickchange(self, muc: tabs.MucTab) -> None:
        "We changed our nick in a MUC"
        for tab in self.core.get_tabs(tabs.PrivateTab):
            if tab.parent_muc == muc:
                tab.own_nick = muc.own_nick

    async def on_groupchat_private_message(self, message: Message, sent: bool) -> None:
        """
        We received a Private Message (from someone in a Muc)
        """
        jid = message['to'] if sent else message['from']
        with_nick = jid.resource
        if not with_nick:
            await self.on_groupchat_message(message)
            return

        room_from = jid.bare
        use_xhtml = config.get_by_tabname(
            'enable_xhtml_im',
            jid.bare
        )
        tmp_dir = get_image_cache()
        body = xhtml.get_body_from_message_stanza(
            message, use_xhtml=use_xhtml, extract_images_to=tmp_dir)
        tab = self.core.tabs.by_name_and_class(
            jid.full,
            tabs.PrivateTab)  # get the tab with the private conversation
        ignore = config.get_by_tabname('ignore_private', room_from)
        if ignore and not sent:
            await self.core.events.trigger_async('ignored_private', message, tab)
            msg = config.get_by_tabname('private_auto_response', room_from)
            if msg and body:
                self.core.xmpp.send_message(
                    mto=jid.full, mbody=msg, mtype='chat')
            return
        if tab and 'reactions' in message:
            tab.handle_reactions(message)

        if tab is None:  # It's the first message we receive: create the tab
            if body and not ignore:
                muctab = self.core.tabs.by_name_and_class(jid.bare, tabs.MucTab)
                tab = tabs.PrivateTab(
                        self.core,
                        jid,
                        self.core.own_nick,
                        initial=message,
                        muctab=muctab,
                )
                self.core.tabs.append(tab)
                if muctab is not None:
                    muctab.privates.append(tab)
        else:
            await tab.handle_message(message)

        if not sent and 'private' in config.getstr('beep_on').split():
            if not config.get_by_tabname('disable_beep', jid.full):
                curses.beep()
        if tab is self.core.tabs.current_tab:
            self.core.refresh_window()
        elif tab is not None:
            tab.state = 'normal' if sent else 'private'
            self.core.refresh_tab_win()

    ### Chatstates ###

    async def on_chatstate_active(self, message: Message) -> None:
        await self._on_chatstate(message, "active")

    async def on_chatstate_inactive(self, message: Message) -> None:
        await self._on_chatstate(message, "inactive")

    async def on_chatstate_composing(self, message: Message) -> None:
        await self._on_chatstate(message, "composing")

    async def on_chatstate_paused(self, message: Message) -> None:
        await self._on_chatstate(message, "paused")

    async def on_chatstate_gone(self, message: Message) -> None:
        await self._on_chatstate(message, "gone")

    async def _on_chatstate(self, message: Message, state: str) -> None:
        if message['type'] == 'chat':
            if not await self._on_chatstate_normal_conversation(message, state):
                tab = self.core.tabs.by_name_and_class(message['from'].full,
                                                       tabs.PrivateTab)
                if not tab:
                    return
                await self._on_chatstate_private_conversation(message, state)
        elif message['type'] == 'groupchat':
            await self.on_chatstate_groupchat_conversation(message, state)

    async def _on_chatstate_normal_conversation(self, message: Message, state: str) -> bool:
        tab = self.core.get_conversation_by_jid(message['from'], False)
        if not tab:
            return False
        await self.core.events.trigger_async('normal_chatstate', message, tab)
        tab.chatstate = state
        if tab == self.core.tabs.current_tab:
            tab.refresh_info_header()
            self.core.doupdate()
        else:
            _composing_tab_state(tab, state)
            self.core.refresh_tab_win()
        return True

    async def _on_chatstate_private_conversation(self, message: Message, state: str) -> None:
        """
        Chatstate received in a private conversation from a MUC
        """
        tab = self.core.tabs.by_name_and_class(message['from'].full,
                                               tabs.PrivateTab)
        if not tab:
            return
        await self.core.events.trigger_async('private_chatstate', message, tab)
        tab.chatstate = state
        if tab == self.core.tabs.current_tab:
            tab.refresh_info_header()
            self.core.doupdate()
        else:
            _composing_tab_state(tab, state)
            self.core.refresh_tab_win()

    async def on_chatstate_groupchat_conversation(self, message: Message, state: str) -> None:
        """
        Chatstate received in a MUC
        """
        nick = message['mucnick']
        room_from = message.get_mucroom()
        tab = self.core.tabs.by_name_and_class(room_from, tabs.MucTab)
        if tab and tab.get_user_by_name(nick):
            await self.core.events.trigger_async('muc_chatstate', message, tab)
            user = tab.get_user_by_name(nick)
            if user:
                user.chatstate = state
        if tab == self.core.tabs.current_tab:
            if not self.core.size.tab_degrade_x:
                tab.user_win.refresh(tab.users)
            tab.input.refresh()
            self.core.doupdate()
        elif tab is not None:
            _composing_tab_state(tab, state)
            self.core.refresh_tab_win()

    @staticmethod
    def _format_error(error: Iq) -> str:
        error_condition = error['condition']
        error_text = error['text']
        return '%s: %s' % (error_condition,
                           error_text) if error_text else error_condition

    def on_version_result(self, iq: Iq) -> None:
        """
        Handle the result of a /version command.
        """
        jid = iq['from']
        if iq['type'] == 'error':
            reply = self._format_error(iq['error'])
            self.core.information(
                'Could not get the software '
                'version from %s: %s' % (jid, reply), 'Warning')
            return
        res = iq['software_version']
        version = '%s is running %s version %s on %s' % (
            jid, res.get('name', 'an unknown software'),
            res.get('version', 'unknown'), res.get('os',
                                                   'an unknown platform'))
        self.core.information(version, 'Info')

    def on_bookmark_result(self, iq: Iq | IqError | IqTimeout | None) -> None:
        """
        Handle the result of a /bookmark commands.
        """
        if iq is None:
            self.core.information(
                'Unable to save bookmarks: no remote storage available.',
                'Warning',
            )
        elif isinstance(iq, (IqError, IqTimeout)):
            real_iq = iq.iq
            reply = self._format_error(real_iq['error'])
            self.core.information(
                'Could not set the remote bookmarks: %s' % reply, 'Warning')
        else:
            self.core.information('Bookmarks saved', 'Info')

    ### subscription-related handlers ###

    async def on_roster_update(self, iq: Iq) -> None:
        """
        The roster was received.
        """
        for item in iq['roster']:
            try:
                jid = item['jid']
            except InvalidJID:
                jid = item._get_attr('jid', '')
                log.exception('Invalid JID: "%s"', jid)
            else:
                if item['subscription'] == 'remove':
                    del roster[jid]
                else:
                    roster.update_contact_groups(jid)
        roster.update_size()
        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()

    async def on_subscription_request(self, presence: Presence) -> None:
        """subscribe received"""
        jid = presence['from'].bare
        contact = roster[jid]
        if contact and contact.subscription in ('from', 'both'):
            return
        elif contact and contact.subscription == 'to':
            self.core.xmpp.send_presence(pto=jid, ptype='subscribed')
            self.core.xmpp.send_presence(pto=jid)
        else:
            if not contact:
                contact = roster.get_and_set(jid)
            roster.update_contact_groups(contact)
            contact.pending_in = True
            self.core.information(
                '%s wants to subscribe to your presence, use '
                '/accept <jid> or /deny <jid> in the roster '
                'tab to accept or reject the query.' % jid, 'Roster')
            self.core.tabs.first().state = 'highlight'
            roster.modified()
        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()

    async def on_subscription_authorized(self, presence: Presence) -> None:
        """subscribed received"""
        jid = presence['from'].bare
        contact = roster[jid]
        if not contact:
            return
        if contact.subscription not in ('both', 'from'):
            self.core.information('%s accepted your contact proposal' % jid,
                                  'Roster')
        if contact.pending_out:
            contact.pending_out = False

        roster.modified()

        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()

    async def on_subscription_remove(self, presence: Presence) -> None:
        """unsubscribe received"""
        jid = presence['from'].bare
        contact = roster[jid]
        if not contact:
            return
        roster.modified()
        self.core.information(
            '%s does not want to receive your status anymore.' % jid, 'Roster')
        self.core.tabs.first().state = 'highlight'
        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()

    async def on_subscription_removed(self, presence: Presence) -> None:
        """unsubscribed received"""
        jid = presence['from'].bare
        contact = roster[jid]
        if not contact:
            return
        roster.modified()
        if contact.pending_out:
            self.core.information('%s rejected your contact proposal' % jid,
                                  'Roster')
            contact.pending_out = False
        else:
            self.core.information(
                '%s does not want you to receive their/its status anymore.' %
                jid, 'Roster')
        self.core.tabs.first().state = 'highlight'
        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()

    ### Presence-related handlers ###

    async def on_presence(self, presence: Presence) -> None:
        if presence.match('presence/muc'):
            return
        jid = presence['from']
        contact = roster[jid.bare]
        if contact is None:
            return
        roster.modified()
        contact.error = None
        await self.core.events.trigger_async('normal_presence', presence,
                                             contact[jid.full])
        tab = self.core.get_conversation_by_jid(jid, create=False)
        if isinstance(tab, tabs.OneToOneTab):
            tab.update_status(
                Status(show=presence['show'], message=presence['status']))
        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()
        elif self.core.tabs.current_tab == tab:
            tab.refresh()
            self.core.doupdate()

    async def on_presence_error(self, presence: Presence) -> None:
        jid = presence['from']
        contact = roster[jid.bare]
        if contact:
            roster.modified()
            contact.error = (
                presence['error']['text'] or
                presence['error']['type'] + ': ' + presence['error']['condition']
            )
            # TODO:  reset chat states status on presence error
            return
        await self.on_groupchat_presence(presence)

    async def on_got_offline(self, presence: Presence) -> None:
        """
        A JID got offline
        """
        if presence.match('presence/muc'):
            return
        jid = presence['from']
        status = presence['status']
        if not logger.log_roster_change(
                jid.bare,
                'got offline{}'.format(' ({})'.format(status) if status else '')
        ):
            self.core.information('Unable to write in the log file', 'Error')
        # If a resource got offline, display the message in the conversation with this
        # precise resource.
        contact = roster[jid.bare]
        name = jid.bare
        if contact:
            roster.connected -= 1
            if contact.name:
                name = contact.name
        offline_msg = '%s is \x191}offline' % name
        if status:
            offline_msg += ' (\x19o%s\x191})' % status
        if jid.resource:
            self.core.add_information_message_to_conversation_tab(
                jid.full, '\x195}' + offline_msg)
        self.core.add_information_message_to_conversation_tab(
            jid.bare, '\x195}' + offline_msg)
        self.core.information('\x193}' + offline_msg,
                              'Roster')
        roster.modified()
        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()

    async def on_got_online(self, presence: Presence) -> None:
        """
        A JID got online
        """
        if presence.match('presence/muc'):
            return
        jid = presence['from']
        contact = roster[jid.bare]
        if contact is None:
            # Todo, handle presence coming from contacts not in roster
            return
        roster.connected += 1
        roster.modified()
        if not logger.log_roster_change(jid.bare, 'got online'):
            self.core.information('Unable to write in the log file', 'Error')
        resource = Resource(
            jid.full, {
                'priority': presence.get_priority() or 0,
                'status': presence['status'],
                'show': presence['show'],
            })
        await self.core.events.trigger_async('normal_presence', presence, resource)
        name = contact.name if contact.name else jid.bare
        self.core.add_information_message_to_conversation_tab(
            jid.full, '\x195}%s is \x194}online' % name)
        if time.time() - self.core.connection_time > 10:
            # We do not display messages if we recently logged in
            if presence['status']:
                self.core.information(
                    "\x193}%s \x195}is \x194}online\x195} (\x19o%s\x195})" %
                    (name, presence['status']), "Roster")
            else:
                self.core.information(
                    "\x193}%s \x195}is \x194}online\x195}" % name, "Roster")
            self.core.add_information_message_to_conversation_tab(
                jid.bare, '\x195}%s is \x194}online' % name)
        if isinstance(self.core.tabs.current_tab, tabs.RosterInfoTab):
            self.core.refresh_window()

    async def on_groupchat_presence(self, presence: Presence) -> None:
        """
        Triggered whenever a presence stanza is received from a user in a multi-user chat room.
        Display the presence on the room window and update the
        presence information of the concerned user
        """
        from_room = presence['from'].bare
        tab = self.core.tabs.by_name_and_class(from_room, tabs.MucTab)
        if tab:
            await self.core.events.trigger_async('muc_presence', presence, tab)
            tab.handle_presence(presence)

    ### Connection-related handlers ###

    async def on_failed_connection(self, error: str) -> None:
        """
        We cannot contact the remote server
        """
        self.core.information(
            "Connection to remote server failed: %s" % (error, ), 'Error')

    async def on_session_end(self, _: Any) -> None:
        """
        Called when a session is terminated (e.g. due to a manual disconnect or a 0198 resume fail)
        """
        self.core.xmpp.roster.reset()
        roster.connected = 0
        roster.modified()
        for tab in self.core.get_tabs(tabs.MucTab):
            tab.disconnect()

    async def on_session_resumed(self, _: Any) -> None:
        """
        Called when a session is successfully resumed by 0198
        """
        self.core.information("Resumed session as %s" % self.core.xmpp.boundjid.full, 'Info')
        self.core.xmpp.plugin['xep_0199'].enable_keepalive()

    async def on_disconnected(self, event: str) -> None:
        """
        When we are disconnected from remote server
        """
        if 'disconnect' in config.getstr('beep_on').split():
            curses.beep()
        # Stop the ping plugin. It would try to send stanza on regular basis
        self.core.xmpp.plugin['xep_0199'].disable_keepalive()
        msg_typ = 'Error' if not self.core.legitimate_disconnect else 'Info'
        self.core.information(
            "Disconnected from server%s." % (event and ": %s" % event or ""),
            msg_typ,
        )
        if self.core.legitimate_disconnect or not config.getbool(
                'auto_reconnect'):
            return
        if (self.core.last_stream_error
                and self.core.last_stream_error[1]['condition'] in (
                    'conflict', 'host-unknown')):
            return
        await asyncio.sleep(1)
        if not self.core.xmpp.is_connecting() and not self.core.xmpp.is_connected():
            self.core.information("Auto-reconnecting.", 'Info')
            self.core.xmpp.start()

    async def on_reconnect_delay(self, event: int) -> None:
        """
        When the reconnection is delayed
        """
        self.core.information("Reconnecting in %d seconds..." % (event), 'Info')

    async def on_stream_error(self, event: StreamError) -> None:
        """
        When we receive a stream error
        """
        if event and event['text']:
            self.core.information('Stream error: %s' % event['text'], 'Error')
        if event:
            self.core.last_stream_error = (time.time(), event)

    async def on_failed_all_auth(self, _: Any) -> None:
        """
        Authentication failed
        """
        self.core.information("Authentication failed (bad credentials?).",
                              'Error')
        self.core.legitimate_disconnect = True

    async def on_no_auth(self, _: Any) -> None:
        """
        Authentication failed (no mech)
        """
        self.core.information(
            "Authentication failed, no login method available.", 'Error')
        self.core.legitimate_disconnect = True

    async def on_connected(self, _: Any) -> None:
        """
        Remote host responded, but we are not yet authenticated
        """
        self.core.information("Connected to server.", 'Info')
        self.core.legitimate_disconnect = False

    async def on_session_start(self, _: Any) -> None:
        """
        Called when we are connected and authenticated
        """
        self.core.connection_time = time.time()
        if not self.core.plugins_autoloaded:  # Do not reload plugins on reconnection
            self.core.autoload_plugins()
        self.core.information("Authentication success.", 'Info')
        self.core.information("Your JID is %s" % self.core.xmpp.boundjid.full,
                              'Info')
        if not self.core.xmpp.anon:
            # request the roster
            self.core.xmpp.get_roster()
            roster.update_contact_groups(self.core.xmpp.boundjid.bare)
            # send initial presence
            if config.getbool('send_initial_presence'):
                pres = self.core.xmpp.make_presence()
                pres['show'] = self.core.status.show
                pres['status'] = self.core.status.message
                await self.core.events.trigger_async('send_normal_presence', pres)
                pres.send()
        self.core.bookmarks.get_local()
        # join all the available bookmarks. As of yet, this is just the local ones
        self.core.join_initial_rooms(self.core.bookmarks.local())

        if config.getbool('enable_user_nick'):
            self.core.xmpp.plugin['xep_0172'].publish_nick(
                nick=self.core.own_nick, callback=dumb_callback)
        self.core.register_task(self.core.xmpp.plugin['xep_0115'].update_caps())
        # Start the ping's plugin regular event
        self.core.xmpp.set_keepalive_values()

    ### Other handlers ###

    async def on_status_codes(self, message: Message) -> None:
        """
        Handle groupchat messages with status codes.
        Those are received when a room configuration change occurs.
        """
        room_from = message['from']
        tab = self.core.tabs.by_name_and_class(room_from, tabs.MucTab)
        status_codes = {
            s.attrib['code']
            for s in message.xml.findall('{%s}x/{%s}status' %
                                         (tabs.NS_MUC_USER, tabs.NS_MUC_USER))
        }
        if '101' in status_codes:
            self.core.information(
                'Your affiliation in the room %s changed' % room_from, 'Info')
        elif tab and status_codes:
            show_unavailable = '102' in status_codes
            hide_unavailable = '103' in status_codes
            non_priv = '104' in status_codes
            logging_on = '170' in status_codes
            logging_off = '171' in status_codes
            non_anon = '172' in status_codes
            semi_anon = '173' in status_codes
            full_anon = '174' in status_codes
            modif = False
            info_col = {'info_col': dump_tuple(get_theme().COLOR_INFORMATION_TEXT)}
            if show_unavailable or hide_unavailable or non_priv or logging_off\
                    or non_anon or semi_anon or full_anon:
                tab.add_message(
                    PersistentInfoMessage(
                        'Info: A configuration change not privacy-related occurred.'
                    ),
                )
                modif = True
            if show_unavailable:
                tab.add_message(
                    PersistentInfoMessage(
                        'Info: The unavailable members are now shown.'
                    ),
                )
            elif hide_unavailable:
                tab.add_message(
                    PersistentInfoMessage(
                        'Info: The unavailable members are now hidden.',
                    ),
                )
            if non_anon:
                tab.add_message(
                    PersistentInfoMessage(
                        '\x191}Warning:\x19%(info_col)s} The room is now '
                        'not anonymous. (public JID)' % info_col
                    ),
                )
            elif semi_anon:
                tab.add_message(
                    PersistentInfoMessage(
                        'Info: The room is now semi-anonymous. '
                        '(moderators-only JID)',
                    ),
                )
            elif full_anon:
                tab.add_message(
                    PersistentInfoMessage(
                        'Info: The room is now fully anonymous.',
                    ),
                )
            if logging_on:
                tab.add_message(
                    PersistentInfoMessage(
                        '\x191}Warning: \x19%(info_col)s}This room is publicly logged' % info_col
                    ),
                )
            elif logging_off:
                tab.add_message(
                    PersistentInfoMessage(
                        'Info: This room is not logged anymore.',
                    ),
                )
            if modif:
                self.core.refresh_window()

    async def on_groupchat_subject(self, message: Message) -> None:
        """
        Triggered when the topic is changed.
        """
        nick_from = message['mucnick']
        room_from = message.get_mucroom()
        tab = self.core.tabs.by_name_and_class(room_from, tabs.MucTab)
        subject = message['subject']
        time = message['delay']['stamp']
        if subject is None or not tab:
            return
        if subject != tab.topic:
            # Do not display the message if the subject did not change or if we
            # receive an empty topic when joining the room.
            theme = get_theme()
            fmt = {
                'info_col': dump_tuple(theme.COLOR_INFORMATION_TEXT),
                'text_col': dump_tuple(theme.COLOR_NORMAL_TEXT),
                'subject': subject,
                'user': '',
                'str_time': time,
            }
            if nick_from:
                user = tab.get_user_by_name(nick_from)
                if nick_from == tab.own_nick:
                    after = ' (You)'
                else:
                    after = ''
                if user:
                    user_col = dump_tuple(user.color)
                    user_string = '\x19%s}%s\x19%s}%s' % (
                        user_col, nick_from, fmt['info_col'], after)
                else:
                    user_string = '\x19%s}%s%s' % (fmt['info_col'], nick_from,
                                                   after)
                fmt['user'] = user_string

            if nick_from:
                tab.add_message(
                    PersistentInfoMessage(
                        "%(user)s set the subject to: \x19%(text_col)s}%(subject)s" % fmt,
                        time=time,
                    ),
                )
            else:
                tab.add_message(
                    PersistentInfoMessage(
                        "The subject is: \x19%(text_col)s}%(subject)s" % fmt,
                        time=time,
                    ),
                )
        tab.topic = subject
        tab.topic_from = nick_from
        if self.core.tabs.by_name_and_class(
                room_from, tabs.MucTab) is self.core.tabs.current_tab:
            self.core.refresh_window()

    async def on_receipt(self, message: Message) -> None:
        """
        When a delivery receipt is received (XEP-0184)
        """
        jid = message['from']
        msg_id = message['receipt']
        if not msg_id:
            return

        conversation = self.core.tabs.by_name_and_class(
            jid.full, tabs.OneToOneTab)
        conversation = conversation or self.core.tabs.by_name_and_class(
            jid.bare, tabs.OneToOneTab)
        if not conversation:
            log.error("Received ack from non-existing chat tab: %s", jid)
            return

        try:
            conversation.ack_message(msg_id, self.core.xmpp.boundjid)
        except AckError:
            log.debug('Error while receiving an ack', exc_info=True)

    async def on_data_form(self, message: Message) -> None:
        """
        When a data form is received
        """
        self.core.information(str(message))

    async def on_attention(self, message: Message) -> None:
        """
        Attention probe received.
        """
        jid_from = message['from']
        self.core.information('%s requests your attention!' % jid_from, 'Info')
        tab = (
            self.core.tabs.by_name_and_class(
                jid_from.full, tabs.ChatTab
            ) or self.core.tabs.by_name_and_class(
                jid_from.bare, tabs.ChatTab
            )
        )
        if tab and tab is not self.core.tabs.current_tab:
            tab.state = "attention"
            self.core.refresh_tab_win()

    def outgoing_stanza(self, stanza: StanzaBase | str | bytes) -> None:
        """
        We are sending a new stanza, write it in the xml buffer if needed.
        """
        if self.core.xml_tab:
            stanza_str = str(stanza)
            if PYGMENTS:
                xhtml_text = highlight(stanza_str, LEXER, FORMATTER)
                poezio_colored = xhtml.xhtml_to_poezio_colors(
                    xhtml_text, force=True).rstrip('\x19o').strip()
            else:
                poezio_colored = stanza_str
            self.core.xml_buffer.add_message(
                XMLLog(txt=poezio_colored, incoming=False),
            )
            try:
                if self.core.xml_tab.match_stanza(
                        ElementBase(ET.fromstring(stanza_str))):
                    self.core.xml_tab.filtered_buffer.add_message(
                        XMLLog(txt=poezio_colored, incoming=False),
                    )
            except Exception:
                # Most of the time what gets logged is whitespace pings. Skip.
                # And also skip tab updates.
                if stanza_str.strip() == '':
                    return None
                log.debug('', exc_info=True)

            if isinstance(self.core.tabs.current_tab, tabs.XMLTab):
                self.core.tabs.current_tab.refresh()
                self.core.doupdate()

    def incoming_stanza(self, stanza: StanzaBase) -> None:
        """
        We are receiving a new stanza, write it in the xml buffer if needed.
        """
        if self.core.xml_tab:
            if PYGMENTS:
                xhtml_text = highlight(str(stanza), LEXER, FORMATTER)
                poezio_colored = xhtml.xhtml_to_poezio_colors(
                    xhtml_text, force=True).rstrip('\x19o').strip()
            else:
                poezio_colored = str(stanza)
            self.core.xml_buffer.add_message(
                XMLLog(txt=poezio_colored, incoming=True),
            )
            try:
                if self.core.xml_tab.match_stanza(stanza):
                    self.core.xml_tab.filtered_buffer.add_message(
                        XMLLog(txt=poezio_colored, incoming=True),
                    )
            except Exception:
                log.debug('', exc_info=True)
            if isinstance(self.core.tabs.current_tab, tabs.XMLTab):
                self.core.tabs.current_tab.refresh()
                self.core.doupdate()

    def ssl_invalid_chain(self, tb: Any) -> None:
        self.core.information('The certificate sent by the server is invalid.',
                              'Error')
        self.core.disconnect()

    def _ssl_pop_tab(self, old_cert: str, new_cert: str) -> None:
        def cb(result: bool) -> None:
            if result:
                self.core.information(
                    'New certificate accepted:\nnew: %s\nold: %s' %
                    (old_cert, new_cert), 'Info')
                log.debug('Setting certificate to %s', new_cert)
                if not config.silent_set('certificate', new_cert):
                    self.core.information('Unable to write in the config file',
                                          'Error')
            else:
                self.core.information(
                    'You refused to validate the certificate.'
                    ' You are now disconnected.', 'Info')
                self.core.disconnect()

        confirm_tab = tabs.ConfirmTab(
            self.core,
            'Certificate check required',
            CERT_WARNING_TEXT % (self.core.xmpp.boundjid.domain, old_cert,
                                 new_cert),
            'You need to accept or reject the certificate',
            cb,
            critical=True)

        self.core.add_tab(confirm_tab, True)
        self.core.doupdate()
        # handle resize
        prev_value = signal.signal(signal.SIGWINCH, self.core.sigwinch_handler)
        while not confirm_tab.done:
            try:
                sel = select.select([sys.stdin], [], [], 0.5)[0]
                if sel:
                    self.core.on_input_readable()
            except Exception:
                continue
        signal.signal(signal.SIGWINCH, prev_value)

    def validate_ssl(self, pem: str) -> None:
        """
        Check the server certificate using the slixmpp ssl_cert event
        """
        if config.getbool('ignore_certificate'):
            return
        cert = config.getstr('certificate')
        # update the cert representation when it uses the old one
        if cert and ':' not in cert:
            cert = ':'.join(
                i + j for i, j in zip(cert[::2], cert[1::2])).upper()
            config.set_and_save('certificate', cert)

        der = ssl.PEM_cert_to_DER_cert(pem)
        asn1 = pyasn1.codec.der.decoder.decode(
            der, asn1Spec=pyasn1_modules.rfc2459.Certificate())[0]
        # pylint: disable=no-member
        spki = asn1.getComponentByName("tbsCertificate").getComponentByName(
            "subjectPublicKeyInfo")
        spki_digest = sha256(
            pyasn1.codec.der.encoder.encode(spki)).hexdigest().upper()
        spki_found_cert = ':'.join(
            i + j for i, j in zip(spki_digest[::2], spki_digest[1::2]))
        sha2_digest = sha512(der).hexdigest().upper()
        sha2_found_cert = ':'.join(
            i + j for i, j in zip(sha2_digest[::2], sha2_digest[1::2]))

        if cert:
            if sha2_found_cert == cert:
                log.debug(
                    'Current hash is cert hash, moving to SPKI hash (%s)',
                    spki_found_cert)
                config.set_and_save('certificate', spki_found_cert)
                return
            elif spki_found_cert == cert:
                return
            else:
                self._ssl_pop_tab(cert, spki_found_cert)
        else:
            log.debug('First time. Setting certificate to %s', spki_found_cert)
            if not config.silent_set('certificate', spki_found_cert):
                self.core.information('Unable to write in the config file',
                                      'Error')

    def http_confirm(self, stanza: StanzaBase) -> None:
        confirm = stanza['confirm']

        def cb(result: bool) -> None:
            if result:
                reply = stanza.reply()
            else:
                reply = stanza.reply()
                reply.enable('error')
                reply['error']['type'] = 'auth'
                reply['error']['code'] = '401'
                reply['error']['condition'] = 'not-authorized'
            reply.append(stanza['confirm'])
            reply.send()

        c_id, c_url, c_method = confirm['id'], confirm['url'], confirm[
            'method']
        confirm_tab = tabs.ConfirmTab(
            self.core,
            'HTTP Verification',
            HTTP_VERIF_TEXT % (c_method, c_url, c_id, stanza['from'].full),
            'An HTTP verification was requested',
            cb,
            critical=False)
        self.core.add_tab(confirm_tab, False)
        self.core.refresh_window()
        self.core.doupdate()

    ### Ad-hoc commands

    def next_adhoc_step(self, iq: Iq, adhoc_session: dict[str, Any]) -> None:
        status = iq['command']['status']
        xform = iq.xml.find(
            '{http://jabber.org/protocol/commands}command/{jabber:x:data}x')
        if xform is not None:
            form = self.core.xmpp.plugin['xep_0004'].build_form(xform)
        else:
            form = None

        if status == 'error':
            self.core.information("An error occurred while executing the command")
            return

        def default_handler(form: Form, session: dict[str, Any]) -> None:
            return
        on_validate = on_cancel = default_handler
        if status == 'executing':
            if not form:
                self.core.information(
                    "Adhoc command step does not contain a data-form. Aborting the execution.",
                    "Error")
                self.core.xmpp.plugin['xep_0050'].cancel_command(
                    adhoc_session)
                return
            on_validate = self._validate_adhoc_step
            on_cancel = self._cancel_adhoc_command
        if status == 'completed':
            def validate_handler(form: Form, session: dict[str, Any]) -> None:
                self.core.close_tab()

            def cancel_handler(form: Form, session: dict[str, Any]) -> None:
                self.core.close_tab()

            on_validate = validate_handler
            on_cancel = cancel_handler

        # If a form is available, use it, and add the Notes from the
        # response to it, if any
        if form:
            for note in iq['command']['notes']:
                form.add_field(type='fixed', label=note[1])
            self.core.open_new_form(
                form, on_cancel, on_validate, session=adhoc_session)
        else:  # otherwise, just display an information
            # message
            notes = '\n'.join([note[1] for note in iq['command']['notes']])
            self.core.information("Adhoc command %s: %s" % (status, notes),
                                  "Info")

    def adhoc_error(self, iq: Iq, adhoc_session: dict[str, Any]) -> None:
        self.core.xmpp.plugin['xep_0050'].terminate_command(adhoc_session)
        error_message = get_error_message(iq)
        self.core.information(
            "An error occurred while executing the command: %s" %
            (error_message), 'Error')

    def _cancel_adhoc_command(self,  form: Form, session: dict[str, Any]) -> None:
        self.core.xmpp.plugin['xep_0050'].cancel_command(session)
        self.core.close_tab()

    def _validate_adhoc_step(self, form: Form, session: dict[str, Any]) -> None:
        session['payload'] = form
        self.core.xmpp.plugin['xep_0050'].continue_command(session)
        self.core.close_tab()

    def _terminate_adhoc_command(self, form: Form, session: dict[str, Any]) -> None:
        self.core.xmpp.plugin['xep_0050'].terminate_command(session)
        self.core.close_tab()


def _composing_tab_state(tab: tabs.Tab, state: str) -> None:
    """
    Set a tab state to or from the "composing" state
    according to the config and the current tab state
    """
    if isinstance(tab, tabs.MucTab):
        values = ['true', 'muc']
    elif isinstance(tab, tabs.PrivateTab):
        values = ['true', 'direct', 'private']
    elif isinstance(tab, tabs.ConversationTab):
        values = ['true', 'direct', 'conversation']
    else:
        return  # should not happen

    show_str = config.getstr('show_composing_tabs').lower()
    show = show_str in values

    if tab.state != 'composing' and state == 'composing':
        if show:
            if tabs.STATE_PRIORITY[tab.state] > tabs.STATE_PRIORITY[state]:
                return
            tab.save_state()
            tab.state = 'composing'
    elif tab.state == 'composing' and state != 'composing':
        tab.restore_state()



"""
``reorder`` plugin: Reorder the tabs according to a layout

Commands
--------

.. glossary::

    /reorder
        **Usage:** ``/reorder``

        Reorder the tabs according to the configuration.

    /save_order
        **Usage:** ``/save_order``

        Save the current tab order to the configuration.

Configuration
-------------

The configuration file must contain a section ``[reorder]`` and each option
must be formatted like ``[tab number] = [tab type]:[tab name]``.

For example:

.. code-block:: ini

    [reorder]
    1 = muc:toto@conference.example.com
    2 = muc:example@muc.example.im
    3 = dynamic:robert@example.org

The ``[tab number]`` must be at least ``1``; if the range is not entirely
covered, e.g.:

.. code-block:: ini

    [reorder]
    1 = muc:toto@conference.example.com
    3 = dynamic:robert@example.org

Poezio will insert gaps between the tabs in order to keep the specified
numbering (so in this case, there will be a tab 1, a tab 3, but no tab 2).


The ``[tab type]`` must be one of:

- ``muc`` (for multi-user chats)
- ``private`` (for chats with a specific user inside a multi-user chat)
- ``dynamic`` (for normal, dynamic conversations tabs)
- ``static`` (for conversations with a specific resource)

And finally, the ``[tab name]`` must be:

- For a type ``muc``, the bare JID of the room
- For a type ``private``, the full JID of the user (room JID with the username as a resource)
- For a type ``dynamic``, the bare JID of the contact
- For a type ``static``, the full JID of the contact
"""

from slixmpp import JID

from poezio import tabs
from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin, PluginConfig
from poezio.config import config
from poezio.core.tabs import Tabs
from typing import Type

TEXT_TO_TAB = {
    'muc': tabs.MucTab,
    'private': tabs.PrivateTab,
    'dynamic': tabs.DynamicConversationTab,
    'static': tabs.StaticConversationTab,
    'empty': tabs.GapTab
}

TAB_TO_TEXT = {
    tabs.MucTab: 'muc',
    tabs.DynamicConversationTab: 'dynamic',
    tabs.PrivateTab: 'private',
    tabs.StaticConversationTab: 'static',
    tabs.GapTab: 'empty'
}


def parse_config(tab_config: PluginConfig) -> dict[int, tuple[Type[tabs.Tab], str]] | None:
    result = {}
    for option in tab_config.options('reorder'):
        if not option.isdecimal():
            continue
        pos = int(option)
        if pos in result or pos <= 0:
            return None

        spec = tab_config.get(option, default=':').split(':', maxsplit=1)
        # Gap tabs are recreated automatically if there's a gap in indices.
        if spec == 'empty':
            return None
        typ, name = spec
        if typ not in TEXT_TO_TAB:
            return None
        result[pos] = (TEXT_TO_TAB[typ], name)

    return result


def check_tab(tab: tabs.Tab) -> str:
    for cls, rep in TAB_TO_TEXT.items():
        if isinstance(tab, cls):
            return rep
    return ''


def parse_runtime_tablist(tablist: Tabs) -> list[tuple[int, str]]:
    props = []
    i = 0
    for tab in tablist[1:]:
        i += 1
        if not isinstance(tab, tabs.ChatTab):
            continue
        result = check_tab(tab)
        # Don't serialize gap tabs as they're recreated automatically
        if result != 'empty' and isinstance(tab, tuple(TEXT_TO_TAB.values())):
            props.append((i, '%s:%s' % (result, tab.jid.full)))
    return props


class Plugin(BasePlugin):
    """reorder plugin"""

    def init(self) -> None:
        self.api.add_command(
            'reorder',
            self.command_reorder,
            help='Reorder all tabs using the pre-defined'
            ' layout from the configuration file.')
        self.api.add_command(
            'save_order',
            self.command_save_order,
            help='Save the current tab layout')

    @command_args_parser.ignored
    def command_save_order(self) -> None:
        """
        /save_order
        """
        conf = parse_runtime_tablist(self.core.tabs)
        for key, value in conf:
            self.config.set(str(key), value)
        self.api.information('Tab order saved', 'Info')

    @command_args_parser.ignored
    def command_reorder(self) -> None:
        """
        /reorder
        """
        tabs_spec = parse_config(self.config)
        if not tabs_spec:
            self.api.information('Invalid reorder config', 'Error')
            return None

        old_tabs = self.core.tabs.get_tabs()
        roster = old_tabs.pop(0)

        create_gaps = config.get('create_gaps')

        new_tabs = [roster]
        last = 0
        for pos in sorted(tabs_spec):
            if create_gaps and pos > last + 1:
                new_tabs += [
                    tabs.GapTab() for i in range(pos - last - 1)
                ]
            cls, jid_str = tabs_spec[pos]
            try:
                jid = JID(jid_str)
                tab = self.core.tabs.by_name_and_class(str(jid), cls=cls)
                if tab and tab in old_tabs:
                    new_tabs.append(tab)
                    old_tabs.remove(tab)
                else:
                    # TODO: Add support for MucTab. Requires nickname.
                    if cls in (tabs.DynamicConversationTab, tabs.StaticConversationTab):
                        self.api.information('Tab %s not found. Creating it' % jid, 'Warning')
                        new_tab = cls(self.core, JID(jid))
                        new_tabs.append(new_tab)
                    else:
                        new_tabs.append(tabs.GapTab())
            except Exception:
                self.api.information('Failed to create tab \'%s\'.' % jid_str, 'Error')
                if create_gaps:
                    new_tabs.append(tabs.GapTab())
            finally:
                last = pos

        for tab in old_tabs:
            if tab:
                new_tabs.append(tab)

        # TODO: Ensure we don't break poezio and call this with whatever
        # tablist we have. The roster tab at least needs to be in there.
        self.core.tabs.replace_tabs(new_tabs)
        self.core.refresh_window()

        return None



"""
Module used to provide fixes for slixmpp functions not yet fixed
upstream.

TODO: Check that they are fixed and remove those hacks
"""

from slixmpp import Message
from slixmpp.xmlstream import StanzaBase
from slixmpp.plugins.xep_0184 import XEP_0184
from typing import TypeVar

import logging

log = logging.getLogger(__name__)

S = TypeVar('S', bound=StanzaBase)


def _filter_add_receipt_request(self: XEP_0184, stanza: S) -> S:
    """
    Auto add receipt requests to outgoing messages, if:

        - ``self.auto_request`` is set to ``True``
        - The message is not for groupchat
        - The message does not contain a receipt acknowledgment
        - The recipient is a bare JID or, if a full JID, one
          that has the ``urn:xmpp:receipts`` feature enabled
        - The message has a body

    The disco cache is checked if a full JID is specified in
    the outgoing message, which may mean a round-trip disco#info
    delay for the first message sent to the JID if entity caps
    are not used.
    """

    if not self.auto_request:
        return stanza

    if not isinstance(stanza, Message):
        return stanza

    if stanza['request_receipt']:
        return stanza

    if stanza['type'] not in self.ack_types:
        return stanza

    if stanza['receipt']:
        return stanza

    if not stanza['body']:
        return stanza

    # hack
    if stanza['to'].resource and not hasattr(stanza, '_add_receipt'):
        return stanza

    stanza['request_receipt'] = True
    return stanza



"""
Windows used inthe bookmarkstab
"""
import curses
from typing import List, Tuple, Optional

from slixmpp import JID, InvalidJID

from poezio.windows.base_wins import Win
from poezio.windows.inputs import Input
from poezio.windows.data_forms import FieldInput, FieldInputWin

from poezio.theming import to_curses_attr, get_theme
from poezio.bookmarks import Bookmark, BookmarkList


class BookmarkInput(FieldInput):
    def __init__(self, field: Bookmark) -> None:
        super().__init__()
        self._field = field


class BookmarkInputWin(FieldInputWin):
    def __init__(self, field: Bookmark) -> None:
        super().__init__()
        self._field = field


class BookmarkNameInput(BookmarkInput, Input):
    def __init__(self, field: Bookmark) -> None:
        BookmarkInput.__init__(self, field)
        Input.__init__(self)
        self.text = str(field.name)
        self.pos = 0
        self.view_pos = 0
        self.color = get_theme().COLOR_NORMAL_TEXT

    def save(self) -> None:
        self._field.name = self.get_text()

    def get_help_message(self) -> str:
        return 'Edit the text'


class BookmarkJIDInput(BookmarkInput, Input):
    def __init__(self, field: Bookmark) -> None:
        BookmarkInput.__init__(self, field)
        Input.__init__(self)
        try:
            jid = JID(field.jid)
        except InvalidJID:
            jid = JID('')
        if field.nick:
            jid.resource = field.nick
        self.text = jid.full
        self.pos = 0
        self.view_pos = 0
        self.color = get_theme().COLOR_NORMAL_TEXT

    def save(self) -> None:
        try:
            jid = JID(self.get_text())
        except InvalidJID:
            jid = JID('')
        self._field.jid = JID(jid.bare)
        self._field.nick = jid.resource

    def get_help_message(self) -> str:
        return 'Edit the text'


class BookmarkMethodInput(BookmarkInputWin):
    def __init__(self, field: Bookmark) -> None:
        BookmarkInputWin.__init__(self, field)
        self.options = ('local', 'remote')
        # val_pos is the position of the currently selected option
        self.val_pos = self.options.index(field.method)

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        if key == 'KEY_LEFT':
            if self.val_pos > 0:
                self.val_pos -= 1
        elif key == 'KEY_RIGHT':
            if self.val_pos < len(self.options) - 1:
                self.val_pos += 1
        else:
            return True
        self.refresh()
        return True

    def refresh(self) -> None:
        self._win.erase()
        self._win.attron(to_curses_attr(self.color))
        self.addnstr(0, 0, ' ' * self.width, self.width)
        if self.val_pos > 0:
            self.addstr(0, 0, '←')
        if self.val_pos < len(self.options) - 1:
            self.addstr(0, self.width - 1, '→')
        if self.options:
            option = self.options[self.val_pos]
            self.addstr(0, self.width // 2 - len(option) // 2, option)
        self._win.attroff(to_curses_attr(self.color))
        self._refresh()

    def save(self) -> None:
        self._field.method = self.options[self.val_pos]

    def get_help_message(self) -> str:
        return '←, →: Select a value amongst the others'


class BookmarkPasswordInput(BookmarkInput, Input):
    def __init__(self, field: Bookmark) -> None:
        BookmarkInput.__init__(self, field)
        Input.__init__(self)
        self.text = field.password or ''
        self.pos = len(self.text)
        self.color = get_theme().COLOR_NORMAL_TEXT

    def rewrite_text(self) -> None:
        self._win.erase()
        if self.color:
            self._win.attron(to_curses_attr(self.color))
        self.addstr(
            '*' * len(self.text[self.view_pos:self.view_pos + self.width - 1]))
        if self.color:
            (y, x) = self._win.getyx()
            size = self.width - x
            self.addnstr(' ' * size, size, to_curses_attr(self.color))
        self.addstr(0, self.pos, '')
        if self.color:
            self._win.attroff(to_curses_attr(self.color))
        self._refresh()

    def save(self) -> None:
        self._field.password = self.get_text() or None

    def get_help_message(self) -> str:
        return 'Edit the secret text'


class BookmarkAutojoinWin(BookmarkInputWin):
    def __init__(self, field: Bookmark) -> None:
        BookmarkInputWin.__init__(self, field)
        self.last_key = 'KEY_RIGHT'
        self.value = field.autojoin

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        if key == 'KEY_LEFT' or key == 'KEY_RIGHT':
            self.value = not self.value
            self.last_key = key
        self.refresh()
        return True

    def refresh(self) -> None:
        self._win.erase()
        self._win.attron(to_curses_attr(self.color))
        format_string = '←{:^%s}→' % 7
        inp = format_string.format(repr(self.value))
        self.addstr(0, 0, inp)
        if self.last_key == 'KEY_RIGHT':
            self.move(0, 8)
        else:
            self.move(0, 0)
        self._win.attroff(to_curses_attr(self.color))
        self._refresh()

    def save(self) -> None:
        self._field.autojoin = self.value

    def get_help_message(self) -> str:
        return '← and →: change the value between True and False'


class BookmarksWin(Win):
    __slots__ = ('scroll_pos', '_current_input', 'current_horizontal_input',
                 '_bookmarks', 'lines')

    def __init__(self, bookmarks: BookmarkList) -> None:
        Win.__init__(self)
        self.scroll_pos = 0
        self._current_input = 0
        self.current_horizontal_input = 0
        self._bookmarks = list(bookmarks)
        self.lines: List[Tuple[BookmarkNameInput, BookmarkJIDInput,
                               BookmarkPasswordInput, BookmarkAutojoinWin,
                               BookmarkMethodInput]] = []
        for bookmark in sorted(self._bookmarks, key=lambda x: str(x.jid)):
            self.lines.append((BookmarkNameInput(bookmark),
                               BookmarkJIDInput(bookmark),
                               BookmarkPasswordInput(bookmark),
                               BookmarkAutojoinWin(bookmark),
                               BookmarkMethodInput(bookmark)))

    @property
    def current_input(self) -> int:
        return self._current_input

    @current_input.setter
    def current_input(self, value: int) -> None:
        if 0 <= self._current_input < len(self.lines):
            if 0 <= value < len(self.lines):
                self.lines[self._current_input][
                    self.current_horizontal_input].set_color(
                        get_theme().COLOR_NORMAL_TEXT)
                self._current_input = value
        else:
            self._current_input = 0

    def add_bookmark(self, bookmark: Bookmark) -> None:
        self.lines.append((BookmarkNameInput(bookmark),
                           BookmarkJIDInput(bookmark),
                           BookmarkPasswordInput(bookmark),
                           BookmarkAutojoinWin(bookmark),
                           BookmarkMethodInput(bookmark)))
        if len(self.lines) > 1:
            self.lines[self.current_input][
                self.current_horizontal_input].set_color(
                    get_theme().COLOR_NORMAL_TEXT)
        self.current_horizontal_input = 0
        if len(self.lines) > 1:
            self.current_input = len(self.lines) - 1
        if self.current_input - self.scroll_pos > self.height - 1:
            self.scroll_pos = self.current_input - self.height + 1
        self.refresh()

    def del_current_bookmark(self) -> Optional[Bookmark]:
        if not self.lines:
            return None
        bm = self.lines[self.current_input][0]._field
        to_delete = self.current_input
        self.current_input -= 1
        del self.lines[to_delete]
        if self.scroll_pos:
            self.scroll_pos -= 1
        self.refresh()
        return bm

    def resize(self, height: int, width: int, y: int, x: int) -> None:
        super().resize(height, width, y, x)
        # Adjust the scroll position, if resizing made the window too small
        # for the cursor to be visible
        while self.current_input - self.scroll_pos > self.height - 1:
            self.scroll_pos += 1

    def go_to_next_line_input(self) -> None:
        if not self.lines:
            return
        if self.current_input == len(self.lines) - 1:
            return
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                get_theme().COLOR_NORMAL_TEXT)
        # Adjust the scroll position if the current_input would be outside
        # of the visible area
        if self.current_input + 1 - self.scroll_pos > self.height - 1:
            self.current_input += 1
            self.scroll_pos += 1
            self.refresh()
        else:
            self.current_input += 1
            self.lines[self.current_input][
                self.current_horizontal_input].set_color(
                    get_theme().COLOR_SELECTED_ROW)

    def go_to_previous_line_input(self) -> None:
        if not self.lines:
            return
        if self.current_input == 0:
            return
        theme = get_theme()
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_NORMAL_TEXT)
        self.current_input -= 1
        # Adjust the scroll position if the current_input would be outside
        # of the visible area
        if self.current_input < self.scroll_pos:
            self.scroll_pos = self.current_input
            self.refresh()
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_SELECTED_ROW)

    def go_to_next_horizontal_input(self) -> None:
        if not self.lines:
            return
        theme = get_theme()
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_NORMAL_TEXT)
        self.current_horizontal_input += 1
        if self.current_horizontal_input > 4:
            self.current_horizontal_input = 0
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_SELECTED_ROW)

    def go_to_next_page(self) -> bool:
        if not self.lines:
            return False

        if self.current_input == len(self.lines) - 1:
            return False

        theme = get_theme()
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_NORMAL_TEXT)
        inc = min(self.height, len(self.lines) - self.current_input - 1)

        if self.current_input + inc - self.scroll_pos > self.height - 1:
            self.current_input += inc
            self.scroll_pos += inc
            self.refresh()
        else:
            self.current_input += inc
            self.lines[self.current_input][
                self.current_horizontal_input].set_color(
                    theme.COLOR_SELECTED_ROW)
        return True

    def go_to_previous_page(self) -> bool:
        if not self.lines:
            return False

        if self.current_input == 0:
            return False

        theme = get_theme()
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_NORMAL_TEXT)

        dec = min(self.height, self.current_input)
        self.current_input -= dec
        # Adjust the scroll position if the current_input would be outside
        # of the visible area
        if self.current_input < self.scroll_pos:
            self.scroll_pos = self.current_input
            self.refresh()
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_SELECTED_ROW)
        return True

    def go_to_previous_horizontal_input(self) -> None:
        if not self.lines:
            return
        if self.current_horizontal_input == 0:
            return
        theme = get_theme()
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_NORMAL_TEXT)
        self.current_horizontal_input -= 1
        self.lines[self.current_input][
            self.current_horizontal_input].set_color(
                theme.COLOR_SELECTED_ROW)

    def on_input(self, key: str, raw: bool = False) -> None:
        if not self.lines:
            return
        self.lines[self.current_input][
            self.current_horizontal_input].do_command(key, raw=raw)

    def refresh(self) -> None:
        # store the cursor status
        self._win.erase()
        y = -self.scroll_pos
        for i in range(len(self.lines)):
            self.lines[i][0].resize(1, self.width // 4, y + 1, 0)
            self.lines[i][1].resize(1, self.width // 4, y + 1, self.width // 4)
            self.lines[i][2].resize(1, self.width // 6, y + 1,
                                    3 * self.width // 6)
            self.lines[i][3].resize(1, self.width // 6, y + 1,
                                    4 * self.width // 6)
            self.lines[i][4].resize(1, self.width // 6, y + 1,
                                    5 * self.width // 6)
            y += 1
        self._refresh()
        for i, inp in enumerate(self.lines):
            if i < self.scroll_pos:
                continue
            if i >= self.height + self.scroll_pos:
                break
            for j in range(5):
                inp[j].refresh()

        if self.lines and self.current_input < self.height - 1:
            self.lines[self.current_input][
                self.current_horizontal_input].set_color(
                    get_theme().COLOR_SELECTED_ROW)
            self.lines[self.current_input][
                self.current_horizontal_input].refresh()
        if not self.lines:
            curses.curs_set(0)
        else:
            curses.curs_set(1)

    def refresh_current_input(self) -> None:
        if self.lines:
            self.lines[self.current_input][
                self.current_horizontal_input].refresh()

    def save(self) -> None:
        for line in self.lines:
            line[0].save()
            line[1].save()
            line[2].save()
            line[3].save()
            line[4].save()



"""
This plugin colors each character of a message with a random color.

.. note:: As ticket `#3273`_ puts it, the final output is closer to vomit than a rainbow.

Usage
-----

.. glossary::

    /rainbow

        Say something in a Chat tab.

.. note:: Can create fun things when used with :ref:`The figlet plugin <figlet-plugin>`.

.. _#3273: https://lab.louiz.org/poezio/poezio/-/issues/3273
"""
from poezio.plugin import BasePlugin
from poezio import xhtml
from slixmpp import Message
from typing import Any

import random

possible_colors = list(range(256))
# remove the colors that are almost white or almost black
for col in [
        16, 232, 233, 234, 235, 236, 237, 15, 231, 255, 254, 253, 252, 251
]:
    possible_colors.remove(col)


def rand_color() -> str:
    return '\x19%s}' % (random.choice(possible_colors), )


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.rainbowize)
        self.api.add_event_handler('private_say', self.rainbowize)
        self.api.add_event_handler('conversation_say', self.rainbowize)

    def rainbowize(self, msg: Message, tab: Any) -> None:
        msg['body'] = ''.join([
            '%s%s' % (
                rand_color(),
                char,
            ) for char in xhtml.clean_text(msg['body'])
        ])



"""
The RosterInfoTab is the tab showing roster info, the list of contacts,
half of it is dedicated to showing the information buffer, and a small
rectangle shows the current contact info.

This module also includes functions to match users in the roster.
"""
import logging
import base64
import curses
import difflib
import os
import ssl
from functools import partial
from os import getenv, path
from pathlib import Path
from typing import Any, Callable, TYPE_CHECKING

from slixmpp import JID, InvalidJID, Message
from slixmpp.exceptions import IqError, IqTimeout

from poezio import windows
from poezio.common import shell_split
from poezio.config import config
from poezio.contact import Contact, Resource
from poezio.decorators import refresh_wrapper
from poezio.roster import RosterGroup, roster
from poezio.decorators import command_args_parser, deny_anonymous
from poezio.core.structs import Command, Completion
from poezio.tabs import DefaultHelpTab, Tab
from poezio.ui.types import InfoMessage

if TYPE_CHECKING:
    from poezio.core.core import Core
    from poezio.windows.inputs import Input

log = logging.getLogger(__name__)


class RosterInfoTab(DefaultHelpTab):
    """
    A tab, split in two, containing the roster and infos
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core') -> None:
        self.default_help_message = windows.HelpText(
            "Enter commands with “/”. “o”: toggle offline show")

        DefaultHelpTab.__init__(self, core)
        self._name = "Roster"
        self.v_separator = windows.VerticalSeparator()
        self.information_win = windows.TextWin()
        self.core.information_buffer.add_window(self.information_win)
        self.roster_win = windows.RosterWin()
        self.contact_info_win = windows.ContactInfoWin()
        self.avatar_win = windows.ImageWin()
        self.state = 'normal'
        self.key_func['^I'] = self.completion
        self.key_func["/"] = self.on_slash
        # disable most of the roster features when in anonymous mode
        if not self.core.xmpp.anon:
            self.key_func[' '] = self.on_space
            self.key_func["KEY_UP"] = self.move_cursor_up
            self.key_func["KEY_DOWN"] = self.move_cursor_down
            self.key_func["M-u"] = self.move_cursor_to_next_contact
            self.key_func["M-y"] = self.move_cursor_to_prev_contact
            self.key_func["M-U"] = self.move_cursor_to_next_group
            self.key_func["M-Y"] = self.move_cursor_to_prev_group
            self.key_func["M-[1;5B"] = self.move_cursor_to_next_group
            self.key_func["M-[1;5A"] = self.move_cursor_to_prev_group
            self.key_func["l"] = lambda: self.command_last_activity("")
            self.key_func["o"] = self.toggle_offline_show
            self.key_func["v"] = self.get_contact_version
            self.key_func["i"] = self.show_contact_info
            self.key_func["s"] = self.start_search
            self.key_func["S"] = self.start_search_slow
            self.key_func["n"] = self.change_contact_name
        self.register_command(
            'name',
            self.command_name,
            usage='<jid> [name]',
            shortdesc='Set the given JID\'s name.',
            completion=self.completion_name)
        self.register_command(
            'groupadd',
            self.command_groupadd,
            usage='[<jid> <group>]|<group>',
            desc='Add the given JID or selected line to the given group.',
            shortdesc='Add a user to a group',
            completion=self.completion_groupadd)
        self.register_command(
            'groupmove',
            self.command_groupmove,
            usage='<jid> <old group> <new group>',
            desc='Move the given JID from the old group to the new group.',
            shortdesc='Move a user to another group.',
            completion=self.completion_groupmove)
        self.register_command(
            'groupremove',
            self.command_groupremove,
            usage='<jid> <group>',
            desc='Remove the given JID from the given group.',
            shortdesc='Remove a user from a group.',
            completion=self.completion_groupremove)
        self.register_command(
            'export',
            self.command_export,
            usage='[/path/to/file]',
            desc='Export your contacts into /path/to/file if '
            'specified, or $HOME/poezio_contacts if not.',
            shortdesc='Export your roster to a file.',
            completion=partial(self.completion_file, 1))
        self.register_command(
            'import',
            self.command_import,
            usage='[/path/to/file]',
            desc='Import your contacts from /path/to/file if '
            'specified, or $HOME/poezio_contacts if not.',
            shortdesc='Import your roster from a file.',
            completion=partial(self.completion_file, 1))
        self.register_command(
            'password',
            self.command_password,
            usage='<password>',
            shortdesc='Change your password')
        self.register_command(
            'disconnect',
            self.command_disconnect,
            desc='Disconnect from the remote server.',
            shortdesc='Disconnect from the server.')
        self.register_command(
            'clear', self.command_clear, shortdesc='Clear the info buffer.')
        self.register_command(
            'last_activity',
            self.command_last_activity,
            usage='<jid>',
            desc='Informs you of the last activity of a JID.',
            shortdesc='Get the activity of someone.',
            completion=self.core.completion.last_activity)

        self.resize()
        self.update_commands()
        self.update_keys()

    def check_blocking(self, features: dict[str, Any]) -> None:
        if 'urn:xmpp:blocking' in features and not self.core.xmpp.anon:
            self.register_command(
                'list_blocks',
                self.command_list_blocks,
                shortdesc='Show the blocked contacts.')
            self.core.xmpp.del_event_handler('session_start',
                                             self.check_blocking)
            self.core.xmpp.add_event_handler('blocked_message',
                                             self.on_blocked_message)

    def check_saslexternal(self, features: dict[str, Any]) -> None:
        if 'urn:xmpp:saslcert:1' in features and not self.core.xmpp.anon:
            self.register_command(
                'certs',
                self.command_certs,
                desc='List the fingerprints of certificates'
                ' which can connect to your account.',
                shortdesc='List allowed client certs.')
            self.register_command(
                'cert_add',
                self.command_cert_add,
                desc='Add a client certificate to the authorized ones. '
                'It must have an unique name and be contained in '
                'a PEM file. [management] is a boolean indicating'
                ' if a client connected using this certificate can'
                ' manage the certificates itself.',
                shortdesc='Add a client certificate.',
                usage='<name> <certificate path> [management]',
                completion=self.completion_cert_add)
            self.register_command(
                'cert_disable',
                self.command_cert_disable,
                desc='Remove a certificate from the list '
                'of allowed ones. Clients currently '
                'using this certificate will not be '
                'forcefully disconnected.',
                shortdesc='Disable a certificate',
                usage='<name>')
            self.register_command(
                'cert_revoke',
                self.command_cert_revoke,
                desc='Remove a certificate from the list '
                'of allowed ones. Clients currently '
                'using this certificate will be '
                'forcefully disconnected.',
                shortdesc='Revoke a certificate',
                usage='<name>')
            self.register_command(
                'cert_fetch',
                self.command_cert_fetch,
                desc='Retrieve a certificate with its '
                'name. It will be stored in <path>.',
                shortdesc='Fetch a certificate',
                usage='<name> <path>',
                completion=self.completion_cert_fetch)

    @property
    def selected_row(self) -> Contact | Resource | RosterGroup | None:
        return self.roster_win.get_selected_row()

    @command_args_parser.ignored
    async def command_certs(self) -> None:
        """
        /certs
        """
        try:
            iq = await self.core.xmpp.plugin['xep_0257'].get_certs(timeout=3)
        except (IqError, IqTimeout):
            self.information(
                'Unable to retrieve the certificate list.', 'Error')
            return
        certs = []
        for item in iq['sasl_certs']['items']:
            users = '\n'.join(item['users'])
            certs.append((item['name'], users))

        if not certs:
            self.information('No certificates found', 'Info')
            return
        msg = 'Certificates:\n'
        msg += '\n'.join(
            (('  %s%s' % (item[0] + (': ' if item[1] else ''), item[1]))
             for item in certs))
        self.information(msg, 'Info')

    @command_args_parser.quoted(2, 1)
    async def command_cert_add(self, args: list[str] | None) -> None:
        """
        /cert_add <name> <certfile> [cert-management]
        """
        if not args or len(args) < 2:
            return self.core.command.help('cert_add')
        name = args[0]

        try:
            with open(args[1]) as fd:
                crt = fd.read()
            crt = crt.replace(ssl.PEM_FOOTER, '').replace(
                ssl.PEM_HEADER, '').replace(' ', '').replace('\n', '')
        except Exception as e:
            self.information(
                'Unable to read the certificate: %s' % e,
                'Error',
            )
            return

        bool_management = True
        if len(args) > 2:
            management = args[2]
            if not management or management.lower() in ('false', '0'):
                bool_management = False

        try:
            await self.core.xmpp.plugin['xep_0257'].add_cert(
                name,
                crt,
                allow_management=bool_management
            )
            self.information('Certificate added.', 'Info')
        except (IqError, IqTimeout):
            self.information(
                'Unable to add the certificate.',
                'Error',
            )

    def completion_cert_add(self, the_input: 'Input') -> Completion | None:
        """
        completion for /cert_add <name> <path> [management]
        """
        n = the_input.get_argument_position()
        log.debug('%s %s %s', the_input.text, n, the_input.pos)
        if n == 1:
            return None
        elif n == 2:
            return self.completion_file(2, the_input)
        elif n == 3:
            return Completion(the_input.new_completion, ['true', 'false'], n)
        return None

    @command_args_parser.quoted(1)
    async def command_cert_disable(self, args: list[str] | None) -> None:
        """
        /cert_disable <name>
        """
        if not args:
            return self.core.command.help('cert_disable')
        name = args[0]
        try:
            await self.core.xmpp.plugin['xep_0257'].disable_cert(name)
            self.information('Certificate disabled.', 'Info')
        except (IqError, IqTimeout):
            self.information(
                'Unable to disable the certificate.',
                'Error',
            )

    @command_args_parser.quoted(1)
    async def command_cert_revoke(self, args: list[str] | None) -> None:
        """
        /cert_revoke <name>
        """
        if not args:
            return self.core.command.help('cert_revoke')
        name = args[0]
        try:
            await self.core.xmpp.plugin['xep_0257'].revoke_cert(name)
            self.information('Certificate revoked.', 'Info')
        except (IqError, IqTimeout):
            self.information(
                'Unable to revoke the certificate.',
                'Error',
            )

    @command_args_parser.quoted(2)
    async def command_cert_fetch(self, args: list[str] | None) -> None:
        """
        /cert_fetch <name> <path>
        """
        if not args or len(args) < 2:
            return self.core.command.help('cert_fetch')
        name = args[0]
        path = args[1]

        try:
            iq = await self.core.xmpp.plugin['xep_0257'].get_certs()
        except (IqError, IqTimeout):
            self.information(
                'Unable to fetch the certificate.',
                'Error',
            )
            return
        bytes_cert = None
        for item in iq['sasl_certs']['items']:
            if item['name'] == name:
                bytes_cert = base64.b64decode(item['x509cert'])
                break
        if not bytes_cert:
            self.information('Certificate not found.', 'Info')
            return
        cert = ssl.DER_cert_to_PEM_cert(bytes_cert)
        with open(path, 'w') as fd:
            fd.write(cert)
        self.information('File stored at %s' % path, 'Info')

    def completion_cert_fetch(self, the_input: 'Input') -> Completion | None:
        """
        completion for /cert_fetch <name> <path>
        """
        n = the_input.get_argument_position()
        log.debug('%s %s %s', the_input.text, n, the_input.pos)
        if n == 1:
            return None
        elif n == 2:
            return self.completion_file(2, the_input)
        return None

    def on_blocked_message(self, message: Message) -> None:
        """
        When we try to send a message to a blocked contact
        """
        tab = self.core.get_conversation_by_jid(message['from'], False)
        if not tab:
            log.debug('Received message from nonexistent tab: %s',
                      message['from'])
            return
        msg = 'Cannot send message to %(jid)s: contact blocked' % {
            'jid': message['from'],
        }
        tab.add_message(InfoMessage(msg))

    @command_args_parser.ignored
    async def command_list_blocks(self) -> None:
        """
        /list_blocks
        """
        try:
            iq = await self.core.xmpp.plugin['xep_0191'].get_blocked()
        except (IqError, IqTimeout):
            self.information(
                'Could not retrieve the blocklist.', 'Error')
            return
        s = 'List of blocked JIDs:\n'
        items = (str(item['jid']) for item in iq['blocklist']['items'])
        jids = '\n'.join(items)
        if jids:
            s += jids
        else:
            s = 'No blocked JIDs.'
        self.information(s, 'Info')

    @command_args_parser.ignored
    def command_disconnect(self) -> None:
        """
        /disconnect
        """
        self.core.disconnect()

    @command_args_parser.quoted(0, 1)
    def command_last_activity(self, args: list[str]) -> None:
        """
        /activity [jid]
        """
        item = self.roster_win.selected_row
        jid: str | JID
        if args:
            jid = args[0]
        elif isinstance(item, Contact):
            jid = item.bare_jid
        elif isinstance(item, Resource):
            jid = item.jid
        else:
            self.information('No JID selected.', 'Error')
            return
        self.core.register_task(
            self.core.command.last_activity(str(jid))
        )

    def resize(self) -> None:
        self.need_resize = False
        if self.size.tab_degrade_x:
            display_info = False
            roster_width = self.width
            display_contact_win = False
            contact_win_h = 0
        else:
            display_info = True
            roster_width = self.width // 2
            if self.size.tab_degrade_y:
                display_contact_win = False
                contact_win_h = 0
            else:
                display_contact_win = True
                contact_win_h = 8
        if self.size.tab_degrade_y:
            tab_win_height = 0
        else:
            tab_win_height = Tab.tab_win_height()

        info_width = self.width - roster_width - 1
        if display_info:
            self.v_separator.resize(self.height - 1 - tab_win_height, 1, 0,
                                    roster_width)
            self.information_win.resize(
                self.height - 1 - tab_win_height - contact_win_h, info_width,
                0, roster_width + 1, self.core.information_buffer,
                force=self.ui_config_changed)
            self.ui_config_changed = False
            if display_contact_win:
                y = self.height - tab_win_height - contact_win_h - 1
                avatar_width = contact_win_h * 2
                self.contact_info_win.resize(contact_win_h,
                                             info_width - avatar_width, y,
                                             roster_width + 1)
                self.avatar_win.resize(contact_win_h, avatar_width, y,
                                       self.width - avatar_width)
        self.roster_win.resize(self.height - 1 - Tab.tab_win_height(),
                               roster_width, 0, 0)
        self.input.resize(1, self.width, self.height - 1, 0)
        self.default_help_message.resize(1, self.width, self.height - 1, 0)

    def completion_file(self, complete_number: int, the_input: 'Input') -> Completion | None:
        """
        Generic quoted completion for files/paths
        (use functools.partial to use directly as a completion
        for a command)
        """
        text = the_input.get_text()
        args = shell_split(text)
        n = the_input.get_argument_position()
        if n == complete_number:
            if args[n - 1] == '' or len(args) < n + 1:
                home = os.getenv('HOME', default='/')
                return Completion(
                    the_input.new_completion, [home, '/tmp'], n, quotify=True)
            path_ = Path(args[n])
            if path_.is_dir():
                dir_ = path_
                base = ''
            else:
                dir_ = path_.parent
                base = path_.name
            try:
                names = list(dir_.iterdir())
            except OSError:
                names = []
            names_filtered = [
                name for name in names if str(name).startswith(base)
            ]
            if names_filtered:
                names = names_filtered
            if not names:
                names = [path_]
            end_list = []
            for name in names:
                if not str(name).startswith('.'):
                    value = dir_ / name
                    end_list.append(str(value))

            return Completion(
                the_input.new_completion, end_list, n, quotify=True)
        return None

    @command_args_parser.ignored
    def command_clear(self) -> None:
        """
        /clear
        """
        self.core.information_buffer.messages = []
        self.information_win.rebuild_everything(self.core.information_buffer)
        self.core.information_win.rebuild_everything(
            self.core.information_buffer)
        self.refresh()

    @deny_anonymous
    @command_args_parser.quoted(1)
    async def command_password(self, args: list[str] | None) -> None:
        """
        /password <password>
        """
        if not args:
            self.information('No password provided.', 'Error')
            return
        try:
            await self.core.xmpp.plugin['xep_0077'].change_password(
                args[0]
            )
            self.information('Password updated', 'Account')
            if config.getstr('password'):
                config.silent_set('password', args[0])
        except (IqError, IqTimeout):
            self.information(
                'Unable to change the password',
                'Account',
            )

    @deny_anonymous
    @command_args_parser.quoted(1, 1)
    async def command_name(self, args: list[str] | None) -> None:
        """
        Set a name for the specified JID in your roster
        """
        if not args:
            return self.core.command.help('name')
        try:
            jid = JID(JID(args[0]).bare)
        except InvalidJID:
            self.information(f'Invalid JID: {args[0]}', 'Error')
            return
        name = args[1] if len(args) == 2 else ''

        contact = roster[jid]
        if contact is None:
            self.information('No such JID in roster', 'Error')
            return

        groups = set(contact.groups)
        if 'none' in groups:
            groups.remove('none')
        subscription = contact.subscription
        try:
            update_roster = self.core.xmpp.update_roster(
                jid,
                name=name,
                groups=groups,
                subscription=subscription
            )
            if update_roster:
                await update_roster
        except (IqError, IqTimeout):
            self.information('The name could not be set.', 'Error')

    @deny_anonymous
    @command_args_parser.quoted(1, 1)
    async def command_groupadd(self, args: list[str] | None) -> None:
        """
        Add the specified JID to the specified group
        """
        if not args:
            return self.core.command.help('groupadd')
        jid: str | JID
        if len(args) == 1:
            group = args[0]
            item = self.roster_win.selected_row
            if isinstance(item, Contact):
                jid = item.bare_jid
            elif isinstance(item, Resource):
                jid = item.jid
            else:
                self.core.command.help('groupadd')
                return
        else:
            try:
                jid = JID(args[0]).bare
            except InvalidJID:
                self.information(f'Invalid JID: {args[0]}', 'Error')
                return
            group = args[1]

        contact = roster[jid]
        if contact is None:
            self.information('No such JID in roster', 'Error')
            return

        new_groups = set(contact.groups)
        if group in new_groups:
            self.information('JID already in group', 'Error')
            return

        roster.modified()
        new_groups.add(group)
        try:
            new_groups.remove('none')
        except KeyError:
            pass

        name = contact.name
        subscription = contact.subscription

        try:
            update_roster = self.core.xmpp.update_roster(
                JID(jid),
                name=name,
                groups=new_groups,
                subscription=subscription,
            )
            if update_roster:
                await update_roster
            roster.update_contact_groups(str(jid))
        except (IqError, IqTimeout):
            self.information('The group could not be set.', 'Error')

    @deny_anonymous
    @command_args_parser.quoted(3)
    async def command_groupmove(self, args: list[str] | None) -> None:
        """
        Remove the specified JID from the first specified group and add it to the second one
        """
        if not args:
            return self.core.command.help('groupmove')
        try:
            jid = JID(args[0]).bare
        except InvalidJID:
            self.information(f'Invalid JID: {args[0]}', 'Error')
            return
        group_from = args[1]
        group_to = args[2]

        contact = roster[jid]
        if not contact:
            self.information('No such JID in roster', 'Error')
            return

        new_groups = set(contact.groups)
        if 'none' in new_groups:
            new_groups.remove('none')

        if group_to == 'none' or group_from == 'none':
            self.information('"none" is not a group.', 'Error')
            return

        if group_from not in new_groups:
            self.information('JID not in first group', 'Error')
            return

        if group_to in new_groups:
            self.information('JID already in second group', 'Error')
            return

        if group_to == group_from:
            self.information('The groups are the same.', 'Error')
            return

        roster.modified()
        new_groups.add(group_to)
        if 'none' in new_groups:
            new_groups.remove('none')

        new_groups.remove(group_from)
        name = contact.name
        subscription = contact.subscription
        try:
            update_roster = self.core.xmpp.update_roster(
                JID(jid),
                name=name,
                groups=new_groups,
                subscription=subscription,
            )
            if update_roster:
                await update_roster
            roster.update_contact_groups(contact)
        except (IqError, IqTimeout):
            self.information('The group could not be set', 'Error')

    @deny_anonymous
    @command_args_parser.quoted(2)
    async def command_groupremove(self, args: list[str] | None) -> None:
        """
        Remove the specified JID from the specified group
        """
        if not args:
            self.core.command.help('groupremove')
            return

        try:
            jid = JID(args[0]).bare
        except InvalidJID:
            self.information(f'Invalid JID: {args[0]}', 'Error')
            return
        group = args[1]

        contact = roster[jid]
        if contact is None:
            self.information('No such JID in roster', 'Error')
            return

        new_groups = set(contact.groups)
        try:
            new_groups.remove('none')
        except KeyError:
            pass
        if group not in new_groups:
            self.information('JID not in group', 'Error')
            return

        roster.modified()

        new_groups.remove(group)
        name = contact.name
        subscription = contact.subscription
        try:
            update_roster = self.core.xmpp.update_roster(
                JID(jid),
                name=name,
                groups=new_groups,
                subscription=subscription,
            )
            if update_roster:
                await update_roster
            roster.update_contact_groups(jid)
        except (IqError, IqTimeout):
            self.information('The group could not be set')

    @deny_anonymous
    @command_args_parser.quoted(0, 1)
    def command_import(self, args: list[str]) -> None:
        """
        Import the contacts
        """
        if args:
            if args[0].startswith('/'):
                filepath = args[0]
            else:
                filepath = path.join(getenv('HOME', ''), args[0])
        else:
            filepath = path.join(getenv('HOME', ''), 'poezio_contacts')
        if not path.isfile(filepath):
            self.information('The file %s does not exist' % filepath,
                             'Error')
            return
        try:
            handle = open(filepath, 'r', encoding='utf-8')
            lines = handle.readlines()
            handle.close()
        except IOError:
            self.information('Could not open %s' % filepath, 'Error')
            log.exception('Unable to correct a message')
            return
        for jid in lines:
            self.core.command.add(jid.lstrip('\n'))
        self.information('Contacts imported from %s' % filepath, 'Info')

    @deny_anonymous
    @command_args_parser.quoted(0, 1)
    def command_export(self, args: list[str]) -> None:
        """
        Export the contacts
        """
        if args:
            if args[0].startswith('/'):
                filepath = args[0]
            else:
                filepath = path.join(getenv('HOME', ''), args[0])
        else:
            filepath = path.join(getenv('HOME', ''), 'poezio_contacts')
        if path.isfile(filepath):
            self.information('The file already exists', 'Error')
            return
        elif not path.isdir(path.dirname(filepath)):
            self.information('Parent directory not found', 'Error')
            return
        if roster.export(filepath):
            self.information('Contacts exported to %s' % filepath, 'Info')
        else:
            self.information(
                'Failed to export contacts to %s' % filepath,
                'Info',
            )

    def completion_remove(self, the_input: 'Input') -> Completion:
        """
        Completion for /remove
        """
        jids = [str(jid) for jid in roster.jids()]
        return Completion(the_input.auto_completion, jids, '', quotify=False)

    def completion_name(self, the_input: 'Input') -> Completion | None:
        """Completion for /name"""
        n = the_input.get_argument_position()
        if n == 1:
            jids = [str(jid) for jid in roster.jids()]
            return Completion(the_input.new_completion, jids, n, quotify=True)
        return None

    def completion_groupadd(self, the_input: 'Input') -> Completion | None:
        n = the_input.get_argument_position()
        if n == 1:
            jids = sorted(str(jid) for jid in roster.jids())
            return Completion(
                the_input.new_completion, jids, n, '', quotify=True)
        elif n == 2:
            groups = sorted(
                group for group in roster.groups if group != 'none')
            return Completion(
                the_input.new_completion, groups, n, '', quotify=True)
        return None

    def completion_groupmove(self, the_input: 'Input') -> Completion | None:
        args = shell_split(the_input.text)
        n = the_input.get_argument_position()
        if n == 1:
            jids = sorted(str(jid) for jid in roster.jids())
            return Completion(
                the_input.new_completion, jids, n, '', quotify=True)
        elif n == 2:
            contact = roster[args[1]]
            if not contact:
                return None
            groups = list(contact.groups)
            if 'none' in groups:
                groups.remove('none')
            return Completion(
                the_input.new_completion, groups, n, '', quotify=True)
        elif n == 3:
            groups = sorted(group for group in roster.groups)
            return Completion(
                the_input.new_completion, groups, n, '', quotify=True)
        return None

    def completion_groupremove(self, the_input: 'Input') -> Completion | None:
        args = shell_split(the_input.text)
        n = the_input.get_argument_position()
        if n == 1:
            jids = sorted(str(jid) for jid in roster.jids())
            return Completion(
                the_input.new_completion, jids, n, '', quotify=True)
        elif n == 2:
            contact = roster[args[1]]
            if contact is None:
                return None
            groups = sorted(contact.groups)
            try:
                groups.remove('none')
            except ValueError:
                pass
            return Completion(
                the_input.new_completion, groups, n, '', quotify=True)
        return None

    def refresh(self) -> None:
        if self.need_resize:
            self.resize()
        log.debug('  TAB   Refresh: %s', self.__class__.__name__)

        display_info = not self.size.tab_degrade_x
        display_contact_win = not self.size.tab_degrade_y

        self.roster_win.refresh(roster)
        if display_info:
            self.v_separator.refresh()
            self.information_win.refresh()
            if display_contact_win:
                row = self.roster_win.get_selected_row()
                if row:
                    self.contact_info_win.refresh(row)
                    if isinstance(row, Contact):
                        self.avatar_win.refresh(row.avatar)
                    else:
                        self.avatar_win.refresh(None)
        self.refresh_tab_win()
        self.input.refresh()

    def on_input(self, key: str, raw: bool) -> None:
        selected_row = None
        if key == '^M':
            selected_row = self.roster_win.get_selected_row()
        res = self.input.do_command(key, raw=raw)
        if res:
            return
        if key == '^M':
            if selected_row:
                self.core.on_roster_enter_key(selected_row)
            return
        elif not raw and key in self.key_func:
            self.key_func[key]()
            return

    @refresh_wrapper.conditional
    def toggle_offline_show(self) -> bool:
        """
        Show or hide offline contacts
        """
        option = 'roster_show_offline'
        value = config.getbool(option)
        success = config.silent_set(option, str(not value))
        roster.modified()
        if not success:
            self.information(
                'Unable to write in the config file',
                'Error',
            )
        return True

    def on_lose_focus(self) -> None:
        self.state = 'normal'

    def on_gain_focus(self) -> None:
        self.state = 'current'
        if isinstance(self.input, windows.HelpText):
            curses.curs_set(0)
        else:
            curses.curs_set(1)

    @refresh_wrapper.conditional
    def move_cursor_down(self) -> bool:
        if isinstance(self.input,
                      windows.Input) and not self.input.history_disabled:
            return False
        return self.roster_win.move_cursor_down()

    @refresh_wrapper.conditional
    def move_cursor_up(self) -> bool:
        if isinstance(self.input,
                      windows.Input) and not self.input.history_disabled:
            return False
        return self.roster_win.move_cursor_up()

    def move_cursor_to_prev_contact(self) -> None:
        self.roster_win.move_cursor_up()
        while not isinstance(self.roster_win.get_selected_row(), Contact):
            if not self.roster_win.move_cursor_up():
                break
        self.roster_win.refresh(roster)

    def move_cursor_to_next_contact(self) -> None:
        self.roster_win.move_cursor_down()
        while not isinstance(self.roster_win.get_selected_row(), Contact):
            if not self.roster_win.move_cursor_down():
                break
        self.roster_win.refresh(roster)

    def move_cursor_to_prev_group(self) -> None:
        self.roster_win.move_cursor_up()
        while not isinstance(self.roster_win.get_selected_row(), RosterGroup):
            if not self.roster_win.move_cursor_up():
                break
        self.roster_win.refresh(roster)

    def move_cursor_to_next_group(self) -> None:
        self.roster_win.move_cursor_down()
        while not isinstance(self.roster_win.get_selected_row(), RosterGroup):
            if not self.roster_win.move_cursor_down():
                break
        self.roster_win.refresh(roster)

    def on_scroll_down(self) -> bool:
        return self.roster_win.move_cursor_down(self.height // 2)

    def on_scroll_up(self) -> bool:
        return self.roster_win.move_cursor_up(self.height // 2)

    @refresh_wrapper.conditional
    def on_space(self) -> bool:
        if isinstance(self.input, windows.Input):
            return False
        selected_row = self.roster_win.get_selected_row()
        if isinstance(selected_row, RosterGroup):
            selected_row.toggle_folded()
            roster.modified()
            return True
        elif isinstance(selected_row, Contact):
            group = "none"
            found_group = False
            pos = self.roster_win.pos
            while not found_group and pos >= 0:
                row = self.roster_win.roster_cache[pos]
                pos -= 1
                if isinstance(row, RosterGroup):
                    found_group = True
                    group = row.name
            selected_row.toggle_folded(group)
            roster.modified()
            return True
        return False

    def get_contact_version(self) -> None:
        """
        Show the versions of the resource(s) currently selected
        """
        selected_row = self.roster_win.get_selected_row()
        if isinstance(selected_row, Contact):
            for resource in selected_row.resources:
                self.core.register_task(
                    self.core.command.version(str(resource.jid))
                )
        elif isinstance(selected_row, Resource):
            self.core.register_task(
                self.core.command.version(str(selected_row.jid))
            )
        else:
            self.information('Nothing to get versions from', 'Info')

    def show_contact_info(self) -> None:
        """
        Show the contact info (resource number, status, presence, etc)
        when 'i' is pressed.
        """
        selected_row = self.roster_win.get_selected_row()
        if isinstance(selected_row, Contact):
            cont = selected_row
            res = selected_row.get_highest_priority_resource()
            acc = []
            acc.append('Contact: %s (%s)' % (cont.bare_jid, res.presence
                                             if res else 'unavailable'))
            if res:
                acc.append(
                    '%s connected resource%s' % (len(cont), ''
                                                 if len(cont) == 1 else 's'))
                acc.append('Current status: %s' % res.status)
            msg = '\n'.join(acc)
        elif isinstance(selected_row, Resource):
            res = selected_row
            msg = 'Resource: %s (%s)\nCurrent status: %s\nPriority: %s' % (
                res.jid, res.presence, res.status, res.priority)
        elif isinstance(selected_row, RosterGroup):
            rg = selected_row
            msg = 'Group: %s [%s/%s] contacts online' % (
                rg.name,
                rg.get_nb_connected_contacts(),
                len(rg),
            )
        else:
            msg = None
        if msg:
            self.information(msg, 'Info')

    def change_contact_name(self) -> None:
        """
        Auto-fill a /name command when 'n' is pressed
        """
        selected_row = self.roster_win.get_selected_row()
        if isinstance(selected_row, Contact):
            jid = selected_row.bare_jid
        elif isinstance(selected_row, Resource):
            jid = JID(JID(selected_row.jid).bare)
        else:
            return
        self.on_slash()
        self.input.text = '/name "%s" ' % jid
        self.input.key_end()
        self.input.refresh()

    @refresh_wrapper.always
    def start_search(self) -> bool:
        """
        Start the search. The input should appear with a short instruction
        in it.
        """
        curses.curs_set(1)
        self.input = windows.CommandInput("[Search]", self.on_search_terminate,
                                          self.on_search_terminate,
                                          self.set_roster_filter)
        self.input.resize(1, self.width, self.height - 1, 0)
        self.input.disable_history()
        roster.modified()
        self.refresh()
        return True

    @refresh_wrapper.always
    def start_search_slow(self) -> bool:
        curses.curs_set(1)
        self.input = windows.CommandInput("[Search]", self.on_search_terminate,
                                          self.on_search_terminate,
                                          self.set_roster_filter_slow)
        self.input.resize(1, self.width, self.height - 1, 0)
        self.input.disable_history()
        return True

    def set_roster_filter_slow(self, txt: str) -> bool:
        roster.contact_filter = (jid_and_name_match_slow, txt)
        roster.modified()
        self.refresh()
        return False

    def set_roster_filter(self, txt: str) -> bool:
        roster.contact_filter = (jid_and_name_match, txt)
        roster.modified()
        self.refresh()
        return False

    @refresh_wrapper.always
    def on_search_terminate(self, txt: str) -> bool:
        curses.curs_set(0)
        roster.contact_filter = roster.DEFAULT_FILTER
        self.reset_help_message()
        roster.modified()
        return True

    def on_close(self) -> None:
        return


def diffmatch(search: str, string: str) -> bool:
    """
    Use difflib and a loop to check if search_pattern can
    be 'almost' found INSIDE a string.
    'almost' being defined by difflib
    """
    if len(search) > len(string):
        return False
    length = len(search)
    ratio = 0.7
    for i in range(len(string) - length + 1):
        if difflib.SequenceMatcher(None, search,
                                   string[i:i + length]).ratio() >= ratio:
            return True
    return False


def jid_and_name_match(contact: Contact, txt: str) -> bool:
    """
    Match jid with text precisely
    """
    if not txt:
        return True
    txt = txt.lower()
    try:
        if txt in JID(contact.bare_jid).bare.lower():
            return True
    except InvalidJID:
        pass
    if txt in contact.name.lower():
        return True
    return False


def jid_and_name_match_slow(contact: Contact, txt: str) -> bool:
    """
    A function used to know if a contact in the roster should
    be shown in the roster
    """
    if not txt:
        return True  # Everything matches when search is empty
    try:
        user = JID(contact.bare_jid).bare
        if diffmatch(txt, user):
            return True
    except InvalidJID:
        pass
    if contact.name and diffmatch(txt, contact.name):
        return True
    return False



"""
The XMLTab is here for debugging purposes, it shows the incoming and
outgoing stanzas. It has a few useful functions that can filter stanzas
in order to only show the relevant ones, and it can also be frozen or
unfrozen on demand so that the relevant information is not drowned by
the traffic.
"""
import curses
import logging
import os
from xml.etree import ElementTree as ET
from typing import TYPE_CHECKING, cast, Callable, Any

from slixmpp import JID, InvalidJID
from slixmpp.xmlstream import matcher, StanzaBase
from slixmpp.xmlstream.stanzabase import ElementBase
from slixmpp.xmlstream.tostring import tostring

from poezio import text_buffer, windows
from poezio.ui.types import XMLLog
from poezio.decorators import command_args_parser
from poezio.tabs import Tab, DefaultHelpTab
from poezio.xhtml import clean_text
from poezio.core.structs import Command


if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)


class MatchJID:
    jid: JID
    dest: str

    def __init__(self, jid: JID, dest: str = '') -> None:
        self.jid = jid
        self.dest = dest

    def match(self, xml: StanzaBase) -> bool:
        from_ = JID(xml['from'])
        to_ = JID(xml['to'])
        if self.jid.full == self.jid.bare:
            from_ = JID(from_.bare)
            to_ = JID(to_.bare)

        if self.dest == 'from':
            return bool(from_ == self.jid)
        elif self.dest == 'to':
            return bool(to_ == self.jid)
        return self.jid in (from_, to_)

    def __repr__(self) -> str:
        return '%s%s%s' % (self.dest, ': ' if self.dest else '', self.jid)


MATCHERS_MAPPINGS = {
    MatchJID: ('JID', repr),
    matcher.MatcherId: ('ID', lambda obj: obj._criteria),
    matcher.MatchXMLMask: ('XMLMask', lambda obj: tostring(obj._criteria)),
    matcher.MatchXPath: ('XPath', lambda obj: obj._criteria)
}


class XMLTab(DefaultHelpTab):
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core') -> None:
        self.default_help_message = windows.HelpText("/ to enter a command")
        DefaultHelpTab.__init__(self, core)
        self.state = 'normal'
        self._name = 'XMLTab'
        self.filters: list[MatchJID | matcher.base.MatcherBase] = []

        self.core_buffer = self.core.xml_buffer
        self.filtered_buffer = text_buffer.TextBuffer()

        self.info_header = windows.XMLInfoWin()
        self.text_win = windows.TextWin()
        self.core_buffer.add_window(self.text_win)

        self.register_command('close', self.close, shortdesc="Close this tab.")
        self.register_command(
            'clear', self.command_clear, shortdesc='Clear the current buffer.')
        self.register_command(
            'filter_reset',
            self.command_filter_reset,
            shortdesc='Reset the stanza filter.')
        self.register_command(
            'filter_id',
            self.command_filter_id,
            usage='<id>',
            desc='Show only the stanzas with the id <id>.',
            shortdesc='Filter by id.')
        self.register_command(
            'filter_xpath',
            self.command_filter_xpath,
            usage='<xpath>',
            desc='Show only the stanzas matching the xpath <xpath>.'
            ' Any occurrences of %n will be replaced by jabber:client.',
            shortdesc='Filter by XPath.')
        self.register_command(
            'filter_jid',
            self.command_filter_jid,
            usage='<jid>',
            desc='Show only the stanzas matching the jid <jid> in from= or to=.',
            shortdesc='Filter by JID.')
        self.register_command(
            'filter_from',
            self.command_filter_from,
            usage='<jid>',
            desc='Show only the stanzas matching the jid <jid> in from=.',
            shortdesc='Filter by JID from.')
        self.register_command(
            'filter_to',
            self.command_filter_to,
            usage='<jid>',
            desc='Show only the stanzas matching the jid <jid> in to=.',
            shortdesc='Filter by JID to.')
        self.register_command(
            'filter_xmlmask',
            self.command_filter_xmlmask,
            usage='<xml mask>',
            desc='Show only the stanzas matching the given xml mask.',
            shortdesc='Filter by xml mask.')
        self.register_command(
            'dump',
            self.command_dump,
            usage='<filename>',
            desc='Writes the content of the XML buffer into a file.',
            shortdesc='Write in a file.')
        self.key_func['^T'] = self.close
        self.key_func['^I'] = self.completion
        self.key_func["KEY_DOWN"] = self.on_scroll_down
        self.key_func["KEY_UP"] = self.on_scroll_up
        self.key_func["^K"] = self.on_freeze
        self.key_func["/"] = self.on_slash
        self.resize()
        # Used to display the infobar
        self.filter_type = ''
        self.filter = ''

    def gen_filter_repr(self) -> None:
        if not self.filters:
            self.filter_type = ''
            self.filter = ''
            return
        filter_types = map(lambda x: MATCHERS_MAPPINGS[type(x)][0],
                           self.filters)
        filter_strings = map(lambda x: MATCHERS_MAPPINGS[type(x)][1](x),
                             self.filters)
        self.filter_type = ','.join(filter_types)
        self.filter = ','.join(filter_strings)

    def update_filters(self, matcher: MatchJID | matcher.base.MatcherBase) -> None:
        if not self.filters:
            messages = self.core_buffer.messages[:]
            self.filtered_buffer.messages = []
            self.core_buffer.del_window(self.text_win)
            self.filtered_buffer.add_window(self.text_win)
        else:
            messages = self.filtered_buffer.messages
            self.filtered_buffer.messages = []
        self.filters.append(matcher)
        new_messages = []
        for msg in messages:
            try:
                if msg.txt.strip() and self.match_stanza(
                        ElementBase(ET.fromstring(clean_text(msg.txt)))):
                    new_messages.append(msg)
            except ET.ParseError:
                log.debug('Malformed XML : %s', msg.txt, exc_info=True)
        self.filtered_buffer.messages = new_messages
        self.text_win.rebuild_everything(self.filtered_buffer)
        self.gen_filter_repr()

    def on_freeze(self) -> None:
        """
        Freeze the display.
        """
        self.text_win.toggle_lock()
        self.refresh()

    def match_stanza(self, stanza: ElementBase) -> bool:
        for matcher_ in self.filters:
            # FIXME: fix the type of matchers in slixmpp, as they do not
            # actually need StanzaBase
            if not matcher_.match(cast(StanzaBase, stanza)):
                return False
        return True

    @command_args_parser.raw
    def command_filter_xmlmask(self, mask: str) -> None:
        """/filter_xmlmask <xml mask>"""
        try:
            self.update_filters(matcher.MatchXMLMask(mask))
            self.refresh()
        except Exception as e:
            self.information('Invalid XML Mask: %s' % e, 'Error')
            self.command_filter_reset("")

    @command_args_parser.raw
    def command_filter_to(self, jid_str: str) -> None:
        """/filter_jid_to <jid>"""
        try:
            jid = JID(jid_str)
        except InvalidJID:
            self.information('Invalid JID: %s' % jid_str, 'Error')
        else:
            self.update_filters(MatchJID(jid, dest='to'))
            self.refresh()

    @command_args_parser.raw
    def command_filter_from(self, jid_str: str) -> None:
        """/filter_jid_from <jid>"""
        try:
            jid = JID(jid_str)
        except InvalidJID:
            self.information('Invalid JID: %s' % jid_str, 'Error')
        else:
            self.update_filters(MatchJID(jid, dest='from'))
            self.refresh()

    @command_args_parser.raw
    def command_filter_jid(self, jid_str: str) -> None:
        """/filter_jid <jid>"""
        try:
            jid = JID(jid_str)
        except InvalidJID:
            self.information('Invalid JID: %s' % jid_str, 'Error')
        else:
            self.update_filters(MatchJID(jid))
            self.refresh()

    @command_args_parser.quoted(1)
    def command_filter_id(self, args: list[str]) -> None:
        """/filter_id <id>"""
        if not args:
            return self.core.command.help('filter_id')

        self.update_filters(matcher.MatcherId(args[0]))
        self.refresh()

    @command_args_parser.raw
    def command_filter_xpath(self, xpath: str) -> None:
        """/filter_xpath <xpath>"""
        try:
            self.update_filters(
                matcher.MatchXPath(
                    xpath.replace('%n', self.core.xmpp.default_ns)))
            self.refresh()
        except Exception:
            self.information('Invalid XML Path', 'Error')
            self.command_filter_reset("")

    @command_args_parser.ignored
    def command_filter_reset(self, ignored: str = '') -> None:
        """/filter_reset"""
        if self.filters:
            self.filters = []
            self.filtered_buffer.del_window(self.text_win)
            self.core_buffer.add_window(self.text_win)
            self.text_win.rebuild_everything(self.core_buffer)
        self.filter_type = ''
        self.filter = ''
        self.refresh()

    @command_args_parser.quoted(1)
    def command_dump(self, args: list[str]) -> None:
        """/dump <filename>"""
        if not args:
            return self.core.command.help('dump')
        if self.filters:
            xml = self.filtered_buffer.messages[:]
        else:
            xml = self.core_buffer.messages[:]
        text = '\n'.join(
            ('%s %s %s' % (
                msg.time.strftime('%H:%M:%S'),
                'IN' if msg.incoming else 'OUT',
                clean_text(msg.txt))
             for msg in xml if isinstance(msg, XMLLog)))
        filename = os.path.expandvars(os.path.expanduser(args[0]))
        try:
            with open(filename, 'w') as fd:
                fd.write(text)
        except Exception as e:
            self.information(
                'Could not write the XML dump: %s' % e,
                'Error',
            )

    def on_scroll_up(self) -> bool:
        return self.text_win.scroll_up(self.text_win.height - 1)

    def on_scroll_down(self) -> bool:
        return self.text_win.scroll_down(self.text_win.height - 1)

    @command_args_parser.ignored
    def command_clear(self, ignored: str = '') -> None:
        """
        /clear
        """
        self.core_buffer.messages = []
        self.filtered_buffer.messages = []
        self.text_win.rebuild_everything(self.filtered_buffer)
        self.refresh()
        self.core.doupdate()

    def close(self, arg: None = None) -> None:
        self.core.close_tab(self)

    def resize(self) -> None:
        self.need_resize = False
        if self.size.tab_degrade_y:
            info_win_size = 0
            tab_win_height = 0
        else:
            info_win_size = self.core.information_win_size
            tab_win_height = Tab.tab_win_height()

        self.text_win.resize(self.height - info_win_size - tab_win_height - 2,
                             self.width, 0, 0)
        self.text_win.rebuild_everything(self.core_buffer)
        self.info_header.resize(
            1, self.width, self.height - 2 - info_win_size - tab_win_height, 0)
        self.input.resize(1, self.width, self.height - 1, 0)

    def refresh(self) -> None:
        if self.need_resize:
            self.resize()
        log.debug('  TAB   Refresh: %s', self.__class__.__name__)

        if self.size.tab_degrade_y:
            display_info_win = False
        else:
            display_info_win = True

        self.text_win.refresh()
        self.info_header.refresh(
            filter_t=self.filter_type,
            filter=self.filter,
            window=self.text_win,
        )
        self.refresh_tab_win()
        if display_info_win:
            self.info_win.refresh()
        self.input.refresh()

    def on_lose_focus(self) -> None:
        self.state = 'normal'

    def on_gain_focus(self) -> None:
        self.state = 'current'
        curses.curs_set(0)

    def on_close(self) -> None:
        super().on_close()
        self.command_clear("")
        self.core.xml_tab = None

    def on_info_win_size_changed(self) -> None:
        if self.core.information_win_size >= self.height - 3:
            return
        self.text_win.resize(
            self.height - 2 - self.core.information_win_size -
            Tab.tab_win_height(), self.width, 0, 0)
        self.info_header.resize(
            1, self.width, self.height - 2 - self.core.information_win_size -
            Tab.tab_win_height(), 0)
